"""
Simulation for REFRAMED models
"""

import logging
from collections import OrderedDict

import numpy as np
from reframed.cobra.simulation import FBA, pFBA, MOMA, lMOMA, ROOM
from reframed.core.cbmodel import CBModel
from reframed.solvers import set_default_solver
from reframed.solvers import solver_instance
from reframed.solvers.solution import Solution
from reframed.solvers.solution import Status as s_status

from . import SimulationMethod, SStatus, get_default_solver
from .simulation import Simulator, SimulationResult, ModelContainer
from ..model.gecko import GeckoModel
from ..util.constants import ModelConstants
from ..util.utilities import elements
from tqdm import tqdm

LOGGER = logging.getLogger(__name__)

solver_map = {'gurobi': 'gurobi', 'cplex': 'cplex', 'glpk': 'optlang'}


# TODO: missing proteins and set objective implementations
class CBModelContainer(ModelContainer):
    """ A basic container for REFRAMED models.

    :param model: A metabolic model.

    """

    def __init__(self, model: CBModel = None):
        self.model = model

    @property
    def id(self):
        return self.model.id

    @property
    def reactions(self):
        return list(self.model.reactions.keys())

    def get_reaction(self, r_id):
        if r_id not in self.reactions:
            raise ValueError(f"Reactions {r_id} does not exist")
        rxn = self.model.reactions[r_id]
        res = {'id': r_id, 'name': rxn.name, 'lb': rxn.lb, 'ub': rxn.ub, 'stoichiometry': rxn.stoichiometry}
        res['gpr'] = str(rxn.gpr) if rxn.gpr is not None else None
        return res

    @property
    def genes(self):
        return list(self.model.genes.keys())

    def get_gene(self, g_id):
        g = self.model.genes[g_id]
        res = {'id': g_id, 'name': g.name}
        return res

    @property
    def metabolites(self):
        return list(self.model.metabolites.keys())

    def get_metabolite(self, m_id):
        met = self.model.metabolites[m_id]
        res = {'id': m_id, 'name': met.name, 'compartment': met.compartment, 'formula': met.metadata.get('FORMULA', '')}
        return res

    @property
    def compartments(self):
        return self.model.compartments

    def get_compartment(self, c_id):
        c = self.model.compartments[c_id]
        res = {'id': c_id, 'name': c.name, 'external': c.external}
        return res

    def get_exchange_reactions(self):
        return self.model.get_exchange_reactions()

    def get_gene_reactions(self):
        """
        :returns: a map of genes to reactions.
        """
        if not self._gene_to_reaction:
            gr = OrderedDict()
            for rxn_id in self.reactions:
                rxn = self.model.reactions[rxn_id]
                if rxn.gpr:
                    genes = rxn.gpr.get_genes()
                    for g in genes:
                        if g in gr.keys():
                            gr[g].append(rxn_id)
                        else:
                            gr[g] = [rxn_id]
            self._gene_to_reaction = gr
        return self._gene_to_reaction

    @property
    def medium(self):

        def is_active(rxn):
            """Determine if a boundary reaction permits flux towards creating
            metabolites
            """
            reaction = self.model.reactions[rxn]
            return ((bool(reaction.get_products()) and (reaction.ub > 0)) or
                    (bool(reaction.get_substrates) and (reaction.lb < 0)))

        def get_active_bound(rxn):
            """For an active boundary reaction, return the relevant bound"""
            reaction = self.model.reactions[rxn]
            if reaction.get_substrates():
                return -reaction.lb
            elif reaction.get_products():
                return reaction.ub

        return {rxn: get_active_bound(rxn) for rxn in self.get_exchange_reactions()
                if is_active(rxn)}


class Simulation(CBModelContainer, Simulator):
    """Generic Simulation class for cobra Model.
       Defines the simulation conditions, and makes available a set of methods.

    :param model: An metabolic model instance.

    Optional:
    :param dic envcond: Dictionary of environmental conditions.
    :param dic constraints: A dictionary of reaction constraints.
    :param solver: An instance of the LP solver.
    :param dic reference: A dictionary of the wild type flux values.

    """

    # TODO: the parent init call is missing ... super() can resolve the mro of the simulation diamond inheritance
    def __init__(self, model: CBModel, envcond=None, constraints=None, solver=None, reference=None,
                 reset_solver=ModelConstants.RESET_SOLVER):

        if not isinstance(model, CBModel):
            raise ValueError(
                "Model is None or is not an instance of REFRAMED CBModel")

        self.model = model
        set_default_solver(solver_map[get_default_solver()])
        # keep track on reaction bounds changes
        self._environmental_conditions = OrderedDict() if envcond is None else envcond
        self._constraints = dict() if constraints is None else {
            k: v for k, v in constraints.items() if k not in list(self._environmental_conditions.keys())}

        self.solver = solver
        self._reference = reference
        self._gene_to_reaction = None
        self.solver = solver
        self._reset_solver = reset_solver
        self.reverse_sintax = [('_b', '_f')]
        self._m_r_lookup = None

        self.__status_mapping = {
            s_status.OPTIMAL: SStatus.OPTIMAL,
            s_status.UNBOUNDED: SStatus.UNBOUNDED,
            s_status.INFEASIBLE: SStatus.INFEASIBLE,
            s_status.INF_OR_UNB: SStatus.INF_OR_UNB,
            s_status.UNKNOWN: SStatus.UNKNOWN,
            s_status.SUBOPTIMAL: SStatus.SUBOPTIMAL
        }
        # apply the env. cond. and additional constraints to the model
        for r_id, bounds in self._environmental_conditions.items():
            self._set_model_reaction_bounds(r_id, bounds)
        for r_id, bounds in self._constraints.items():
            self._set_model_reaction_bounds(r_id, bounds)

    def _set_model_reaction_bounds(self, r_id, bounds):
        if isinstance(bounds, tuple):
            lb = bounds[0]
            ub = bounds[1]
        elif isinstance(bounds, (int, float)):
            lb = bounds
            ub = bounds
        else:
            raise ValueError(f"Invalid bounds definition {bounds}")
        self.model.set_flux_bounds(r_id, lb, ub)

    @property
    def environmental_conditions(self):
        return self._environmental_conditions.copy()

    @environmental_conditions.setter
    def environmental_conditions(self, a):
        raise ValueError("Can not change environmental conditions. Use set_reaction_bounds instead")

    @property
    def objective(self):
        return self.model.get_objective()

    @objective.setter
    def objective(self, objective):
        a = self.model.get_objective()
        d = {k: 0 for k in a}
        if isinstance(objective, str):
            d[objective] = 1
        elif isinstance(objective, dict):
            d.update(objective)
        else:
            raise ValueError(
                'The objective must be a reaction identifier or a dictionary of \
                reaction identifier with respective coeficients.')

        self.model.set_objective(d)

    def update(self):
        """Updates the model
        """
        self.model.update()

    def add_metabolite(self, id, formula=None, name=None, compartment=None):
        """Adds a metabolite

        :param id: [description]
        :type id: [type]
        :param formula: [description], defaults to None
        :type formula: [type], optional
        :param name: [description], defaults to None
        :type name: [type], optional
        :param compartment: [description], defaults to None
        :type compartment: [type], optional
        """
        from reframed.core.model import Metabolite
        meta = Metabolite(id, name=name, compartment=compartment)
        meta.metadata['FORMULA'] = formula
        self.model.add_metabolite(meta)

    def add_reaction(self, rxn_id, stoichiometry, lb=ModelConstants.REACTION_LOWER_BOUND,
                     ub=ModelConstants.REACTION_UPPER_BOUND, replace=True, **kwargs):
        """Adds a reaction to the model

        Args:
            rxn_id: The reaction identifier
            stoichiometry: The reaction stoichiometry, a dictionary of species: coefficient
            replace (bool, optional): If the reaction should be replaced in case it is already defined.\
            Defaults to True.
        """
        from reframed.core.cbmodel import CBReaction
        reaction = CBReaction(rxn_id, stoichiometry=stoichiometry, lb=lb, ub=ub)
        self.model.add_reaction(reaction, replace=replace)

    def remove_reaction(self, r_id):
        """Removes a reaction from the model.

        Args:
            r_id (str): The reaction identifier.
        """
        self.model.remove_reaction(r_id)

    def get_uptake_reactions(self):
        """
        :returns: The list of uptake reactions.

        """
        drains = self.get_exchange_reactions()
        reacs = [r for r in drains if self.model.reactions[r].reversible or
                 ((self.model.reactions[r].lb is None or self.model.reactions[r].lb < 0)
                  and len(self.model.reactions[r].get_substrates()) > 0) or
                 ((self.model.reactions[r].ub is None or self.model.reactions[r].ub > 0)
                  and len(self.model.reactions[r].get_products())) > 0]
        return reacs

    def get_transport_reactions(self):
        """
        :returns: The list of transport reactions.
        """
        transport_reactions = []
        for rx in self.reactions:
            s_set = set()
            p_set = set()
            s = self.model.reactions[rx].get_substrates()
            for x in s:
                c = self.model.metabolites[x].compartment
                s_set.add(c)
            p = self.model.reactions[rx].get_products()
            for x in p:
                c = self.model.metabolites[x].compartment
                p_set.add(c)
            if len(s) == 1 and len(p) == 1 and len(p_set.intersection(s_set)) == 0:
                transport_reactions.append(rx)
        return transport_reactions

    def get_transport_genes(self):
        """Returns the list of genes that only catalyze transport reactions.
        """
        trp_rxs = self.get_transport_reactions()
        r_g = self.get_gene_reactions()
        genes = []
        for g, rxs in r_g.items():
            if set(rxs).issubset(set(trp_rxs)):
                genes.append(g)
        return genes

    def reverse_reaction(self, reaction_id):
        """
        Identify if a reaction is reversible and returns the
        reverse reaction if it is the case.

        :param reaction_id: A reaction identifier.
        :return: A reverse reaction identifier or None

        """

        # TODO: ... use regex instead.

        rxn = self.model.reactions[reaction_id]
        reactions = self.model.reactions
        if rxn.lb < 0:
            rxn.reversible = True
            return reaction_id
        # The model might have been converted to irreversible by REFRAMED in which case reversible reactions
        # are decoupled into forward (reaction_id+'_f') and backward (reaction_id+'_b') reactions
        # or migth be using some other identifier which must be included in self.reverse_sufix
        else:
            for a, b in self.reverse_sintax:
                n = len(reaction_id) - len(a)
                m = len(reaction_id) - len(b)
                if reaction_id[n:] == a and reactions[reaction_id[:n] + b]:
                    return reaction_id[:n] + b
                elif reaction_id[m:] == b and reactions[reaction_id[:m] + a]:
                    return reaction_id[:m] + a
                else:
                    continue
            return None

    def metabolite_reaction_lookup(self, force_recalculate=False):
        """ Return the network topology as a nested map from metabolite to reaction to coefficient.
        :return: a dictionary lookup table
        """

        if not self._m_r_lookup or force_recalculate:
            self._m_r_lookup = OrderedDict([(m_id, OrderedDict()) for m_id in self.metabolites])

            for r_id, reaction in self.model.reactions.items():
                for m_id, coeff in reaction.stoichiometry.items():
                    self._m_r_lookup[m_id][r_id] = coeff

        return self._m_r_lookup

    def metabolite_elements(self, metabolite_id):
        formula = self.model.metabolites[metabolite_id].metadata['FORMULA']
        return elements(formula)

    def get_reaction_bounds(self, reaction):
        """
        Returns the bounds for a given reaction.
        :param reaction: str, reaction ID
        :return: lb(s), ub(s), tuple
        """
        lb, ub = self.model.reactions[reaction].lb, self.model.reactions[reaction].ub
        return lb if lb > -np.inf else ModelConstants.REACTION_LOWER_BOUND,\
            ub if ub < np.inf else ModelConstants.REACTION_UPPER_BOUND

    def set_reaction_bounds(self, reaction, lb=None, ub=None):
        """
        Sets the bounds for a given reaction.
        :param reaction: str, reaction ID
        :param float lb: lower bound 
        :param float ub: upper bound
        """
        if rxn in self.get_uptake_reactions():
            self._environmental_conditions[rxn] = (lb, ub)
        else:
            self._constraints[rxn] = (lb, ub)
        self.model.set_flux_bounds(reaction, lb, ub)

    def find_bounds(self):
        """
        Return the median upper and lower bound of the metabolic model.
        Bounds can vary from model to model. Cobrapy defaults to (-1000, 1000).
        """
        lower_bounds = np.asarray([rxn.lb for rxn in self.model.reactions], dtype=float)
        upper_bounds = np.asarray([rxn.ub for rxn in self.model.reactions], dtype=float)
        lower_bound = np.nanmedian(lower_bounds[lower_bounds != 0.0])
        upper_bound = np.nanmedian(upper_bounds[upper_bounds != 0.0])
        if np.isnan(lower_bound):
            LOGGER.warning("Could not identify a median lower bound.")
            lower_bound = ModelConstants.REACTION_LOWER_BOUND
        if np.isnan(upper_bound):
            LOGGER.warning("Could not identify a median upper bound.")
            upper_bound = ModelConstants.REACTION_UPPER_BOUND
        return lower_bound, upper_bound

    def find_unconstrained_reactions(self):
        """Return list of reactions that are not constrained at all."""
        lower_bound, upper_bound = self.find_bounds()
        return [
            rxn
            for rxn in self.model.reactions
            if rxn.lb <= lower_bound and rxn.ub >= upper_bound
        ]

    # Simulate
    def simulate(self, objective=None, method=SimulationMethod.FBA,
                 maximize=True, constraints=None, reference=None,
                 scalefactor=None, solver=None, slim=False):
        '''
        Simulates a phenotype when applying a set constraints using the specified method.

        :param dic objective: The simulation objective. If none, the model objective is considered.
        :param method: The SimulationMethod (FBA, pFBA, lMOMA, etc ...)
        :param boolean maximize: The optimization direction
        :param dic constraints: A dictionary of constraints to be applied to the model.
        :param dic reference: A dictionary of reaction flux values.
        :param float scalefactor: A positive scaling factor for the solver. Default None.
        :param solver: An instance of the solver.
        '''

        if not objective:
            objective = self.model.get_objective()

        simul_constraints = OrderedDict()
        if constraints:
            simul_constraints.update({k: v for k, v in constraints.items()
                                      if k not in list(self._environmental_conditions.keys())})

        a_solver = solver
        if not self._reset_solver and not a_solver:
            if self.solver is None:
                self.solver = solver_instance(self.model)
            a_solver = self.solver

        # scales the model if a scalling factor is defined.
        # ... scalling should be implemented at the solver level.
        if scalefactor:
            for _, rxn in self.model.reactions.items():
                rxn.lb = rxn.lb * scalefactor
                rxn.ub = rxn.ub * scalefactor
            if simul_constraints:
                for idx, constraint in simul_constraints.items():
                    if isinstance(constraint, (int, float)):
                        simul_constraints[idx] = constraint * scalefactor
                    elif isinstance(constraint, tuple):
                        simul_constraints[idx] = tuple(
                            x * scalefactor for x in constraint)
                    else:
                        raise ValueError("Could not scale the model")

        # TODO: I have implemented a single dispatch for this, so we can avoid coding long if else chains.
        #  If it is too "obscure", a dictionary can be used, or a function like get_simulation_method.
        #  Either way, I believe it would be maintainable to have a hidden method for each simulation method

        # TODO: simplifly ...
        if method in [SimulationMethod.lMOMA, SimulationMethod.MOMA, SimulationMethod.ROOM] and reference is None:
            reference = self.reference

        if method == SimulationMethod.FBA:
            get_values = not slim
            solution = FBA(self.model, objective=objective, minimize=not maximize,
                           constraints=simul_constraints, solver=a_solver, get_values=get_values)
        elif method == SimulationMethod.pFBA:
            solution = pFBA(self.model, objective=objective, minimize=not maximize,
                            constraints=simul_constraints, solver=a_solver, obj_frac=0.999)
        elif method == SimulationMethod.lMOMA:
            solution = lMOMA(self.model, constraints=simul_constraints,
                             reference=reference, solver=a_solver)
        elif method == SimulationMethod.MOMA:
            solution = MOMA(self.model, constraints=simul_constraints,
                            reference=reference, solver=a_solver)
        elif method == SimulationMethod.ROOM:
            solution = ROOM(self.model, constraints=simul_constraints,
                            reference=reference, solver=a_solver)
        # Special case in which only the simulation context is required without any simulatin result
        elif method == SimulationMethod.NONE:
            solution = Solution(status=s_status.UNKNOWN,
                                message=None, fobj=None, values=None)
        else:
            raise Exception(
                "Unknown method to perform the simulation.")

        # undoes the model scaling
        if scalefactor:
            for _, rxn in self.model.reactions.items():
                rxn.lb = rxn.lb / scalefactor
                rxn.ub = rxn.ub / scalefactor
            if solution.status in (s_status.OPTIMAL, s_status.SUBOPTIMAL):
                solution.fobj = solution.fobj / scalefactor
                for x, y in solution.values.items():
                    solution.values[x] = y / scalefactor

        if slim:
            return solution.fobj

        else:
            status = self.__status_mapping[solution.status]
            result = SimulationResult(self.model, solution.fobj, fluxes=solution.values, status=status,
                                      envcond=self.environmental_conditions, model_constraints=self._constraints.copy(),
                                      simul_constraints=constraints, maximize=maximize, method=method)
            return result

    def FVA(self, obj_frac=0.9, reactions=None, constraints=None, loopless=False, internal=None, solver=None,
            format='dict'):
        """ Flux Variability Analysis (FVA).

        :param model: An instance of a constraint-based model.
        :param float obj_frac: The minimum fraction of the maximum growth rate (default 0.9).\
            Requires that the objective value is at least the fraction times maximum objective value.\
            A value of 0.85 for instance means that the objective has to be at least at 85% percent of its maximum.
        :param list reactions: List of reactions to analyze (default: all).
        :param dic constraints: Additional constraints (optional).
        :param boolean loopless: Run looplessFBA internally (very slow) (default: false).
        :param list internal: List of internal reactions for looplessFBA (optional).
        :param solver: A pre-instantiated solver instance (optional)
        :param format: The return format: 'dict', returns a dictionary,'df' returns a data frame.
        :returns: A dictionary of flux variation ranges.

        """
        simul_constraints = {}
        if constraints:
            simul_constraints.update({k: v for k, v in constraints.items()
                                      if k not in list(self._environmental_conditions.keys())})

        from reframed.cobra.variability import FVA
        res = FVA(self.model, obj_frac=obj_frac, reactions=reactions,
                  constraints=simul_constraints, loopless=loopless, internal=internal, solver=solver)
        if format == 'df':
            import pandas as pd
            e = res.items()
            f = [[a, b, c] for a, [b, c] in e]
            df = pd.DataFrame(f, columns=['Reaction ID', 'Minimum', 'Maximum'])
            return df
        return res


class GeckoSimulation(Simulation):

    def __init__(self, model: GeckoModel, envcond=None, constraints=None, solver=None, reference=None,
                 reset_solver=ModelConstants.RESET_SOLVER, protein_prefix=None):
        super(GeckoSimulation, self).__init__(
            model, envcond, constraints, solver, reference, reset_solver)
        self.protein_prefix = protein_prefix if protein_prefix else 'draw_prot_'
        self._essential_proteins = None

    @property
    def proteins(self):
        return self.model.proteins

    @property
    def protein_rev_reactions(self):
        return self.model.protein_rev_reactions

    def essential_proteins(self, min_growth=0.01):
        if self._essential_proteins is not None:
            return self._essential_proteins
        wt_solution = self.simulate()
        wt_growth = wt_solution.objective_value
        self._essential_proteins = []
        proteins = self.model.proteins
        for p in tqdm(proteins):
            rxn = "{}{}".format(self.protein_prefix, p)
            res = self.simulate(constraints={rxn: 0})
            if res:
                if (res.status == SStatus.OPTIMAL and res.objective_value < wt_growth * min_growth) \
                        or res.status == SStatus.INFEASIBLE:
                    self._essential_proteins.append(rxn)
        return self._essential_proteins

    def protein_reactions(self, protein):
        """
        Returns the list of reactions associated to a protein
        """
        reactions = []
        for r_id, rxn in self.model.reactions.items():
            lsub = rxn.get_substrates()
            for m in lsub:
                if protein in m:
                    reactions.append(r_id)
        return reactions

    def reverse_reaction(self, reaction_id):
        """
        Identify if a reaction is reversible and returns the
        reverse reaction if it is the case.

        :returns: A reaction identifier or None.
        """
        f, d = zip(*self.model.protein_rev_reactions.values())
        if reaction_id in f:
            return d[f.index(reaction_id)]
        elif reaction_id in d:
            return f[d.index(reaction_id)]
        else:
            return None

    def getKcat(self, protein):
        """ Returns a dictionary of reactions and respective Kcat for a specific enzyme·
        """
        m_r = self.metabolite_reaction_lookup()
        r_d = m_r[protein]
        return {k: v for k, v in r_d.items() if v < 0}



import logging
import itertools

from .problem import AbstractKOProblem, AbstractOUProblem
from ..util.parsing import GeneEvaluator, build_tree, Boolean
from ..simulation import SStatus

logger = logging.getLogger(__name__)


def gene_has_associated_enzyme(model, gene_id):
    if any([gene_id in x.composition for x in model.enzymes]):
        try:
            return model._get_translation_name(gene_id)
        except Exception:
            return None
    return None


def associated_enzyme(model, gene):
    return [x for x in model.enzymes if gene in x.composition]


class ETFLGKOProblem(AbstractKOProblem):
    """
    Gene Knockout Optimization Problem.

    :param model: The constraint metabolic model.
    :param list fevaluation: A list of callable EvaluationFunctions.

    Optional:

    :param OrderedDict envcond: Environmental conditions.
    :param OrderedDict constraints: Additional constraints to be applied to the model.
    :param int candidate_min_size: The candidate minimum size (Default EAConstants.MIN_SOLUTION_SIZE)
    :param int candidate_max_size: The candidate maximum size (Default EAConstants.MAX_SOLUTION_SIZE)
    :param list target: List of modification target genes.
    :param list non_target: List of non target genes. Not considered if a target list is provided.
    :param float scalefactor: A scaling factor to be used in the LP formulation.
    :param boolean only_gpr: Only uses GPRs and do not alter pseudo translation reactions bounds (Default False).

    """

    def __init__(self, model, fevaluation=None, **kwargs):
        super(ETFLGKOProblem, self).__init__(
            model, fevaluation=fevaluation, **kwargs)
        self._only_gpr = kwargs.get('only_gpr', False)
        self.gene_reaction_mapping()

    def gene_reaction_mapping(self):
        """ Maps genes with associated enzymes to reactions."""
        enzyme_reaction = {}
        for rx in self.model.reactions:
            if hasattr(rx, 'enzymes') and rx.enzymes:
                for e in rx.enzymes:
                    enzyme_reaction.setdefault(e.id, []).append(rx.id)
        gene_reaction = {}
        self.has_enzyme = []
        for g in self.simulator.genes:
            gene_reaction[g] = []
            ee = associated_enzyme(self.model, g)
            if ee:
                self.has_enzyme.append(g)
                for e in ee:
                    gene_reaction[g].extend(enzyme_reaction[e.id])
        self.gene_enzyme_reaction = gene_reaction

    def _build_target_list(self):
        print("Building modification target list.")
        genes = set(self.simulator.genes)
        # GPR-based
        print("Computing essential genes.")
        essential = set(self.simulator.essential_genes())
        transport = set(self.simulator.get_transport_genes())
        target = genes - essential - transport
        if self.non_target:
            target = target - set(self.non_target)
        target = list(target)
        self._trg_list = target

    def solution_to_constraints(self, candidate):
        """
        Converts a candidate, dict of genes:0 into a dictionary of constraints.
        """
        genes = list(candidate.keys())
        gr_constraints = dict()
        no_trans = []
        # Translation
        for g in genes:
            if g in self.has_enzyme:
                try:
                    rx = self.model._get_translation_name(g)
                    gr_constraints[rx] = 0
                except Exception:
                    no_trans.append(g)
        # GPR based reaction KO
        active_genes = set(self.simulator.genes) - set(no_trans)
        active_reactions = self.simulator.evaluate_gprs(active_genes)
        # reactions for which there are enzymes whose gene translation has been KO
        #catalyzed_reactions = set(itertools.chain.from_iterable(
        #    [self.gene_enzyme_reaction[g] for g in genes]))
        inactive_reactions = set(self.simulator.reactions) - set(active_reactions)  # - catalyzed_reactions
        gr_constraints.update({rxn: 0 for rxn in inactive_reactions})
        return gr_constraints


class ETFLGOUProblem(AbstractOUProblem):
    """Gene Over/Under expression Optimization Problem

    :param model: The constraint metabolic model.
    :param list fevaluation: A list of callable EvaluationFunctions.

    Optional:

    :param OrderedDict envcond: Environmental conditions.
    :param OrderedDict constraints: Additional constraints to be applied to the model.
    :param int candidate_min_size: The candidate minimum size (Default EAConstants.MIN_SOLUTION_SIZE)
    :param int candidate_max_size: The candidate maximum size (Default EAConstants.MAX_SOLUTION_SIZE)
    :param list target: List of modification target genes.
    :param list non_target: List of non target genes. Not considered if a target list is provided.
    :param float scalefactor: A scaling factor to be used in the LP formulation.
    :param dic reference: Dictionary of flux values to be used in the over/under expression values computation.
    :param tuple operators: (and, or) operations. Default (MIN, MAX).
    :param list levels: Over/under expression levels (Default EAConstants.LEVELS).
    :param boolean twostep: If deletions should be applied before identifiying reference flux values.
    :param boolean only_gpr: Only uses GPRs and do not alter pseudo translation reactions bounds (Default False).

    Note:  Operators that can not be pickled may be defined by a string e.g. 'lambda x,y: (x+y)/2'.


    Up and down regulations are applied on E(T)FL models following a multi-step strategy:
    1) If a gene has an associated enzyme, the gene translation pseudo-reaction has its bounds altered, reflecting
    the modification on expression;
    2) Genes that do not have associated enzymes, have their expression altered using reactions GPRs 

    """

    def __init__(self, model, fevaluation=None, **kwargs):
        super(ETFLGOUProblem, self).__init__(
            model, fevaluation=fevaluation, **kwargs)
        # operators to replace 'and'/'or'. By default min/max
        self._temp_op = kwargs.get('operators', None)
        self._operators = None
        self._only_gpr = kwargs.get('only_gpr', False)
        self.gene_reaction_mapping()

    def gene_reaction_mapping(self):
        # map enzyme to reactions
        enzyme_reaction = {}
        for rx in self.model.reactions:
            if hasattr(rx, 'enzymes') and rx.enzymes:
                for e in rx.enzymes:
                    enzyme_reaction.setdefault(e.id, []).append(rx.id)
        gene_reaction = {}
        self.has_enzyme = []
        for g in self.simulator.genes:
            gene_reaction[g] = []
            ee = associated_enzyme(self.model, g)
            if ee:
                self.has_enzyme.append(g)
            for e in ee:
                gene_reaction[g].extend(enzyme_reaction[e.id])
        self.gene_enzyme_reaction = gene_reaction

    def _build_target_list(self):
        print("Building modification target list.")
        genes = set(self.simulator.genes)
        transport = set(self.simulator.get_transport_genes())
        target = genes - transport
        if self.non_target:
            target = target - set(self.non_target)
        if self._partial_solution:
            target = target - set(self._partial_solution.keys())
        target = list(target)
        self._trg_list = target

    def __op(self):
        # set default operators as configurable options
        if self._operators:
            pass
        elif not self._temp_op or None in self._temp_op or len(self._temp_op) < 2:
            self._operators = (lambda x, y: min(x, y), lambda x, y: max(x, y))
        else:
            ops = []
            for i in [0, 1]:
                op = None
                if isinstance(self._temp_op[i], str):
                    op = eval(self._temp_op[i])
                else:
                    op = self._temp_op[i]
                if callable(op):
                    ops.append(op)
                else:
                    raise ValueError(f"The operator at index {i} is not callable.")
            self._operators = tuple(ops)

    def __deletions(self, candidate):
        """
        Converts deletions to constraints. This method is used to infer reference values
        for under and over regulation.
        """
        genes = list(candidate.keys())
        gr_constraints = dict()
        no_trans = []
        # Translation
        for g in genes:
            if g in self.has_enzyme:
                try:
                    rx = self.model._get_translation_name(g)
                    gr_constraints[rx] = 0
                except Exception:
                    no_trans.append(g)
        # GPR based reaction KO
        active_genes = set(self.simulator.genes) - set(no_trans)
        active_reactions = self.simulator.evaluate_gprs(active_genes)
        #catalyzed_reactions = set(itertools.chain.from_iterable(
        #    [self.gene_enzyme_reaction[g] for g in genes if g not in no_trans]))
        inactive_reactions = set(self.simulator.reactions) - set(active_reactions)  # - catalyzed_reactions
        gr_constraints.update({rxn: 0 for rxn in inactive_reactions})
        return gr_constraints

    def solution_to_constraints(self, candidate):
        """
        Decodes a candidate, a dict of genes:lv into a dictionary of reaction constraints
        """
        gr_constraints = dict()
        reference = self.reference
        if self.twostep:
            try:
                deletions = {rxn: lv for rxn, lv in candidate.items() if lv == 0}
                constr = self.__deletions(deletions)
                sr = self.simulator.simulate(constraints=constr, method='pFBA')
                if sr.status in (SStatus.OPTIMAL, SStatus.SUBOPTIMAL):
                    reference = sr.fluxes
            except Exception as e:
                logger.warning(f"{candidate}: {e}")

        
        no_trans = []
        # translation reactions
        for gene_id, lv in candidate.items():
            if gene_id in self.has_enzyme:
                try:
                    rx = self.model._get_translation_name(gene_id)
                    gr_constraints.update(
                        self.reaction_constraints(rx, lv, reference))
                except Exception:
                    no_trans.append(gene_id)
        catalyzed_reactions = set(itertools.chain.from_iterable(
            [self.gene_enzyme_reaction[g] for g in candidate]))
        # GPR based reaction
        self.__op()
        # evaluate gpr
        evaluator = GeneEvaluator(
            candidate, self._operators[0], self._operators[1])
        for rxn_id in self.simulator.reactions:
            if rxn_id not in catalyzed_reactions:
                gpr = self.simulator.get_gpr(rxn_id)
                if gpr:
                    tree = build_tree(gpr, Boolean)
                    # apply the operators to obtain a level for the reaction
                    # if a gene as no level associated its factor is 1 (see GeneEvaluator)
                    lv = tree.evaluate(evaluator.f_operand, evaluator.f_operator)
                    # debugging
                    # logger.debug(f"{gpr}\n{genes}\nlevel:{lv}\n")

                    # adds the reaction constraint
                    rev_rxn = self.simulator.reverse_reaction(rxn_id)
                    # skips if the reverse reaction was already processed
                    if rev_rxn and rev_rxn in gr_constraints.keys():
                        continue
                    elif lv < 0:
                        raise ValueError("All UO levels should be positive")
                    else:
                        gr_constraints.update(
                            self.reaction_constraints(rxn_id, lv, reference))
        return gr_constraints



import os
from collections import OrderedDict
from time import time
from math import inf
import hashlib
from copy import deepcopy

from reframed.io.sbml import load_cbmodel, parse_gpr_rule
from reframed.core.model import AttrOrderedDict, ReactionType, Compartment, Metabolite
from reframed.core.cbmodel import CBModel, CBReaction, Gene

from .problem import AbstractKOProblem
from mewpy.simulation import get_simulator
from mewpy.simulation.simulation import Simulator


class CommunityKOProblem(AbstractKOProblem):
    """
    Community Knockout Optimization Problem.

    :param models: A list of metabolic models.
    :param list fevaluation: A list of callable EvaluationFunctions.

    Optional parameters:

    :param OrderedDict envcond: Environmental conditions.
    :param OrderedDict constraints: Additional constraints to be applied to the model.
    :param int candidate_min_size: The candidate minimum size (Default EAConstants.MIN_SOLUTION_SIZE)
    :param int candidate_max_size: The candidate maximum size (Default EAConstants.MAX_SOLUTION_SIZE)
    :param list target: List of modification target reactions.
    :param list non_target: List of non target reactions. Not considered if a target list is provided.
    :param float scalefactor: A scaling factor to be used in the LP formulation.
    
    """

    def __init__(self, models: list, fevaluation=[], copy_models=True, **kwargs):
        super(CommunityKOProblem, self).__init__(
            None, fevaluation=fevaluation, **kwargs)
        self.organisms = AttrOrderedDict()
        self.model_ids = list({model.id for model in models})

        if len(self.model_ids) < len(models):
            warn("Model ids are not unique, repeated models will be discarded.")

        for model in models:

            m = model if isinstance(model, Simulator) else get_simulator(model)
            self.organisms[m.id] = deepcopy(m) if copy_models else m

    def merge_models(self, models: list = None):
        """Merges the models. Adapted from REFRAMED 

        :param models: [description], defaults to None.
        :type models: list, optional
        :return: [description]
        :rtype: [type]
        """
        if models is None:
            mlist = self.model_ids
        else:
            mlist = models
        sid = ' '.join(sorted(mlist))
        id = hashlib.md5(sid.encode())
        # TODO: use the id to identify communities already evaluated

        comm_model = CBModel(id)
        old_ext_comps = []
        ext_mets = []

        # default IDs
        ext_comp_id = "ext"
        biomass_id = "community_biomass"
        comm_growth = "community_growth"

        # create external compartment

        comp = Compartment(
            ext_comp_id, "extracellular environment", external=True)
        comm_model.add_compartment(comp)

        # community biomass

        met = Metabolite(biomass_id, "Total community biomass", ext_comp_id)
        comm_model.add_metabolite(met)

        rxn = CBReaction(comm_growth, name="Community growth rate",
                         reversible=False, stoichiometry={biomass_id: -1},
                         lb=0, ub=inf, objective=1)

        comm_model.add_reaction(rxn)

        # add each organism

        for org_id, model in self.organisms.items():

            def rename(old_id):
                return f"{old_id}_{org_id}"

            # add internal compartments

            for c_id in model.compartments:
                comp = model.get_compartment(c_id)
                if comp['external']:
                    old_ext_comps.append(c_id)
                else:
                    new_comp = Compartment(rename(c_id), comp['name'])
                    comm_model.add_compartment(new_comp)

            # add metabolites

            for m_id in model.metabolites:
                met = model.get_metabolite(m_id)
                if met['compartment'] not in old_ext_comps:  # if is internal
                    new_id = rename(m_id)
                    new_met = Metabolite(
                        new_id, met['name'], rename(met['compartment']))
                    new_met.metadata['FORMULA'] = met['formula']
                    comm_model.add_metabolite(new_met)

                elif m_id not in comm_model.metabolites:  # if is external but was not added yet
                    new_met = Metabolite(m_id, met['name'], ext_comp_id)
                    new_met.metadata['FORMULA'] = met['formula']
                    comm_model.add_metabolite(new_met)
                    ext_mets.append(new_met.id)

            # add genes

            for g_id in model.genes:
                gene = model.get_gene(g_id)
                new_id = rename(g_id)
                new_gene = Gene(new_id, gene['name'])
                comm_model.add_gene(new_gene)

            # add internal reactions
            ex_rxns = model.get_exchange_reactions()
            for r_id in model.reactions:
                rxn = model.get_reaction(r_id)
                if r_id in ex_rxns:
                    continue

                new_id = rename(r_id)
                new_stoichiometry = {
                    m_id if m_id in ext_mets else rename(m_id): coeff
                    for m_id, coeff in rxn['stoichiometry'].items()
                }

                if r_id in [x for x, v in model.objective.items() if v > 0]:
                    new_stoichiometry[biomass_id] = 1

                if rxn['gpr'] is None:
                    new_gpr = None
                else:
                    new_gpr = parse_gpr_rule(rxn['gpr'])

                new_rxn = CBReaction(
                    new_id,
                    name=rxn['name'],
                    stoichiometry=new_stoichiometry,
                    lb=rxn['lb'],
                    ub=rxn['ub'],
                    gpr_association=new_gpr
                )

                comm_model.add_reaction(new_rxn)

        # Add exchange reactions

        for m_id in ext_mets:
            r_id = f"R_EX_{m_id[2:]}" if m_id.startswith(
                "M_") else f"R_EX_{m_id}"
            rxn = CBReaction(r_id, reversible=True, stoichiometry={m_id: -1},
                             reaction_type=ReactionType.EXCHANGE)
            comm_model.add_reaction(rxn)
        return comm_model

    def _build_target_list(self):
        """Target organims, i.e., organisms that may be removed from the community.
        """
        print("Building modification target list.")
        target = set(self.model_ids)
        if self.non_target is not None:
            target = target - set(self.non_target)
        self._trg_list = list(target)

    def solution_to_constraints(self, candidate):
        """Returns a model not a list of constraints.

        :param candidate: [description]
        :return: [description]
        """
        ko_organisms = list(candidate.keys())
        models = [x for x in self.model_ids if x not in ko_organisms]
        cmodel = self.merge_models(models)
        return cmodel

    def evaluate_solution(self, solution, decode=True):
        """
        Evaluates a single solution, a community.

        :param solution: The solution to be evaluated (a community model).
        :param decode: If the solution needs to be decoded (convert a list of model ids to a community model).
        :returns: A list of fitness values.
        """
        decoded = {}
        # decoded constraints
        if decode:
            decoded = self.decode(solution)
            cmodel = self.solution_to_constraints(decoded)
        else:
            cmodel = solution
        try:
            p = []
            for method in self.methods:
                sim = get_simulator(cmodel, self.environmental_conditions)
                simulation_result = self.simulator.simulate(method=method, scalefactor=self.scalefactor)
                simulation_results[method] = simulation_result
            # apply the evaluation function(s)
            for f in self.fevaluation:
                v = f(simulation_results, decoded,
                      scalefactor=self.scalefactor)
                p.append(v)
        except Exception as e:
            p = []
            for f in self.fevaluation:
                p.append(f.worst_fitness)
        return p



from typing import Type, Union, TYPE_CHECKING

from .builder import Builder
from .engines import Engines

if TYPE_CHECKING:

    from mewpy.io.engines.engine import Engine

    from io import TextIOWrapper

    try:
        # noinspection PyPackageRequirements
        from cobra import Model as Cobra_Model

    except ImportError:
        Cobra_Model = str

    try:
        # noinspection PyPackageRequirements
        from reframed import CBModel as Reframed_Model

    except ImportError:

        Reframed_Model = str


class Reader(Builder):

    # The Reader is just a wrapper for an engine. It just wraps configurations and provides the simple read API
    # Yet, this object is the user interface to read files, file handlers, cbm models from cobrapy and reframed

    # Wrapping is accomplished using the composition pattern

    # The only difference between the builders and the engines is that engines read the model in five steps:
    #             - open
    #             - parse
    #             - read
    #             - close
    #             - clean

    # Builders, on the other hand, just read and thus performing all operations at once.
    # This is important for reading multiple files into a single model.
    #
    # For reading files in stages, multiple readers must be created and the director must be used to merge all the
    # readers into a single model

    # noinspection PyShadowingNames
    def __init__(self,
                 engine: Type['Engine'],
                 io: Union[str, 'TextIOWrapper', 'Cobra_Model', 'Reframed_Model'],
                 sep: str = ',',
                 id_col: int = 0,
                 target_col: int = 0,
                 regulator_col: int = 1,
                 rule_col: int = 1,
                 co_activating_col: int = 1,
                 co_repressing_col: int = 2,
                 aliases_cols: Union[int, tuple, list] = None,
                 header: Union[None, int] = None,
                 filter_nan: bool = False,
                 config: dict = None):

        """
        Read a given file type (or model type) into a mewpy model (metabolic, regulatory, metabolic-regulatory).

        It currently supports xml, sbml, json, csv and txt. It also supports parsing of cobrapy and reframed models.

        An engine listed in the Engines enumerator of the mewpy.io.engines module must be provided. This engine
        determines the model type.

        :type engine: Engine
        :type io: Union[str, CobraModel, CBModel]

        :param engine: A valid Engine listed in the Engines enumerator of the mewpy.io.engines.
        It will handle file parsing and model building
        :param io: A valid string path or IO is acceptable.
        Alternatively, it can also be provided a cobrapy and reframed model
        :param sep: column separator
        :param id_col: identifier column
        :param target_col: target column
        :param regulator_col: regulator column
        :param rule_col: regulatory rule column
        :param co_activating_col: activating regulators column. Regulators must be separated by spaces
        :param co_repressing_col: repressing regulators column. Regulators must be separated by spaces
        :param aliases_cols: aliases multiple columns
        :param header: If there is an header, please provide the integer for the header row
        :param filter_nan: filter regulatory interactions without explicit regulatory rule
        :param config: dictionary with additional configurations
        """

        if not engine:
            raise ValueError('Nothing to read. Please provide an engine')

        if not io:
            raise ValueError('Nothing to read. Please provide a path, file handler or model')

        if not isinstance(engine, Engines):

            old_engine = engine

            engine = Engines.get(old_engine)

            if engine is None:
                raise ValueError(f'{old_engine} is not supported. See available engines at {Engines}')

        engine: 'Engine' = engine.value

        if not aliases_cols:
            aliases_cols = []

        if not config:
            config = {}

        params_config = dict(sep=sep,
                             id_col=id_col,
                             target_col=target_col,
                             rule_col=rule_col,
                             co_activating_col=co_activating_col,
                             co_repressing_col=co_repressing_col,
                             aliases_cols=aliases_cols,
                             header=header,
                             filter_nan=filter_nan)

        config.update(params_config)

        super(Reader, self).__init__()

        self._engine = engine(io=io, config=config)

    # reading with contexts

    def __enter__(self):

        self._context = True

        self.engine.open()
        self.engine.parse()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):

        self._context = False

        self.engine.close()
        self.engine.clean()

    # to understand this read method, please consult the director, builder and reader init stubs

    def read(self, warnings: bool = True):

        """
        Reading a mewpy model, namely metabolic, regulatory or both encoded into one file type
        (e.g. sbml, csv, cobrapy, reframed, json, etc).
        The file is closed upon reading or failure

        :param warnings: Whether to launch warnings found during reading
        """

        if self._context:

            model = self.engine.read()

        else:
            self.engine.open()
            self.engine.parse()

            model = self.engine.read()

            self.engine.close()
            self.engine.clean()

        if warnings:

            for warning in self.warnings:
                warning()

        model.clean_history()

        return model



from collections import defaultdict
from typing import Union, TYPE_CHECKING, List, Dict, Tuple

# TODO: this module depends on pandas dataframes. Should it be set as package requirement?
from pandas import DataFrame

from mewpy.util.constants import ModelConstants
from mewpy.mew.variables import Regulator, Reaction, Gene
from .rfba import RFBA
from .srfba import SRFBA
from .analysis_utils import decode_solver_solution

if TYPE_CHECKING:
    from mewpy.model import Model, MetabolicModel, RegulatoryModel


# TODO: missing documentation and typing
def _rfba(method, objective, minimize, initial_state, constraints, status=False):

    # noinspection PyProtectedMember
    old_linear_obj = method._linear_objective.copy()
    # noinspection PyProtectedMember
    old_quadratic_obj = method._quadratic_objective.copy()
    old_sense = method.minimize

    sol = method.optimize(objective=objective,
                          minimize=minimize,
                          initial_state=initial_state,
                          constraints=constraints,
                          to_solver=True,
                          get_values=False)

    method.set_objective(linear=old_linear_obj, quadratic=old_quadratic_obj, minimize=old_sense)

    return decode_solver_solution(solution=sol, minimize=minimize, status=status)


def slim_rfba(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
              rfba: RFBA = None,
              objective: Union[str, Dict[str, Union[float, int]]] = None,
              minimize: bool = False,
              initial_state: Dict[str, Union[float, int]] = None,
              constraints: Dict[str, Tuple[Union[float, int], Union[float, int]]] = None) -> Union[int, float, None]:

    if not rfba:
        rfba = RFBA(model, build=True, attach=False)

    return _rfba(method=rfba,
                 objective=objective,
                 minimize=minimize,
                 initial_state=initial_state,
                 constraints=constraints)


def slim_srfba(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
               srfba: SRFBA = None,
               objective: Union[str, Dict[str, Union[float, int]]] = None,
               minimize: bool = False,
               initial_state: Dict[str, Union[float, int]] = None,
               constraints: Dict[str, Tuple[Union[float, int], Union[float, int]]] = None) -> Union[int, float, None]:
    if not srfba:
        srfba = SRFBA(model, build=True, attach=False)

    return _rfba(method=srfba,
                 objective=objective,
                 minimize=minimize,
                 initial_state=initial_state,
                 constraints=constraints)


def _inputs_processing(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None,
                       method: str = None,
                       fraction: float = None,
                       reactions: List[Union[str, Reaction]] = None,
                       genes: List[Union[str, Gene]] = None,
                       regulators: List[Union[str, Regulator]] = None,
                       objective: Union[str, Dict[str, Union[float, int]]] = None,
                       initial_state: Dict[str, Union[float, int]] = None,
                       constraints: Dict[str, Tuple[Union[float, int], Union[float, int]]] = None):
    methods_map = {'rfba': RFBA,
                   'srfba': SRFBA}

    if not constraints:
        constraints = {}

    if not initial_state:
        initial_state = {}

    if objective:

        if isinstance(objective, str):

            rxn = model.get(objective)

            if not rxn:
                ValueError(f'The {objective} objective is not available in the {model} model')

            objective = {objective: 1.0}

        elif isinstance(objective, dict):

            initial_obj = objective

            objective = {}

            for key in initial_obj:

                if isinstance(key, tuple):
                    ValueError('Quadratic objectives are not allowed')

                rxn = model.get(key)

                if not rxn:
                    ValueError(f'The {key} objective is not available in the {model} model')

        else:
            raise TypeError(f'Objective wrong type {type(objective)}')

    else:
        objective = {rxn.id: coef for rxn, coef in model.objective.items()}

        if len(objective) > 1:
            raise ValueError('linear objectives having more than one variable or quadratic objectives are not allowed')

    obj_rxn = next(iter(objective.keys()))

    _RFBA = methods_map.get(method, None)

    if _RFBA is None:
        raise ValueError(f'Simulation method {method} is not valid. Available methods: {methods_map}')

    rfba = _RFBA(model, build=True, attach=False)
    rfba.set_objective(linear=objective, minimize=False)

    if fraction is not None:

        if fraction > 0.0:
            sol = _rfba(method=rfba,
                        objective=objective,
                        minimize=False,
                        initial_state=initial_state,
                        constraints=constraints,
                        status=False)

            if sol is None:
                sol = 0.0

            constraints[obj_rxn] = (fraction * sol, ModelConstants.REACTION_UPPER_BOUND)

    if not reactions:
        reactions = model.reactions.keys()

    if not genes:
        genes = model.genes.keys()

    if not regulators:
        regulators = model.regulators.keys()

    return rfba, reactions, genes, regulators, initial_state, constraints


def ifva(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
         method: str = 'srfba',
         fraction: float = 1.0,
         reactions: List[Union[str, Reaction]] = None,
         objective: Union[str, Dict[str, Union[float, int]]] = None,
         initial_state: Dict[str, Union[float, int]] = None,
         constraints: Dict[str, Tuple[Union[float, int], Union[float, int]]] = None,
         to_dict: bool = False) -> Union[Dict[str, List[Union[float, int]]], DataFrame]:
    """

    Integrated Flux Variability Analysis (iFVA)

    :param model:
    :param method:
    :param fraction:
    :param reactions:
    :param objective:
    :param initial_state:
    :param constraints:
    :param to_dict:

    :return:
    """

    rfba, reactions, _, _, initial_state, constraints = _inputs_processing(model=model,
                                                                           method=method,
                                                                           fraction=fraction,
                                                                           reactions=reactions,
                                                                           genes=None,
                                                                           regulators=None,
                                                                           objective=objective,
                                                                           initial_state=initial_state,
                                                                           constraints=constraints)

    res = defaultdict(list)

    for rxn in reactions:

        if isinstance(rxn, str):

            rxn_objective = rxn

        else:
            rxn_objective = rxn.id

        sol = _rfba(method=rfba,
                    objective=rxn_objective,
                    minimize=True,
                    initial_state=initial_state,
                    constraints=constraints)

        res[rxn_objective].append(sol)

        sol = _rfba(method=rfba,
                    objective=rxn_objective,
                    minimize=False,
                    initial_state=initial_state,
                    constraints=constraints)

        res[rxn_objective].append(sol)

    if to_dict:
        return res

    return DataFrame.from_dict(data=res, orient='index', columns=['minimum', 'maximum'])


def isingle_gene_deletion(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                          method: str = 'srfba',
                          genes: List[Union[str, Gene]] = None,
                          objective: Union[str, Dict[str, Union[float, int]]] = None,
                          initial_state: Dict[str, Union[float, int]] = None,
                          constraints: Dict[str, Tuple[Union[float, int], Union[float, int]]] = None,
                          to_dict: bool = False) -> Union[Dict[str, List[Union[float, int]]], DataFrame]:
    """

    Integrated single gene deletions

    :param model:
    :param method:
    :param genes:
    :param objective:
    :param initial_state:
    :param constraints:
    :param to_dict:

    :return:
    """
    rfba, _, genes, _, initial_state, constraints = _inputs_processing(model=model,
                                                                       method=method,
                                                                       fraction=None,
                                                                       reactions=None,
                                                                       genes=genes,
                                                                       regulators=None,
                                                                       objective=objective,
                                                                       initial_state=initial_state,
                                                                       constraints=constraints)

    res = {}

    for gene in genes:

        if isinstance(gene, str):

            gene_id = gene

        else:
            gene_id = gene.id

        initial_state[gene_id] = 0.0

        sol, status = _rfba(method=rfba,
                            objective=None,
                            minimize=False,
                            initial_state=initial_state,
                            constraints=constraints,
                            status=True)

        res[gene_id] = [sol, status]

        del initial_state[gene_id]

    if to_dict:
        return res

    return DataFrame.from_dict(data=res, orient='index', columns=['growth', 'status'])


def isingle_reaction_deletion(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                              method: str = 'srfba',
                              reactions: List[Union[str, Reaction]] = None,
                              objective: Union[str, Dict[str, Union[float, int]]] = None,
                              initial_state: Dict[str, Union[float, int]] = None,
                              constraints: Dict[str, Tuple[Union[float, int], Union[float, int]]] = None,
                              to_dict: bool = False) -> Union[Dict[str, List[Union[float, int]]], DataFrame]:
    """

    Integrated single reaction deletions

    :param model:
    :param method:
    :param reactions:
    :param objective:
    :param initial_state:
    :param constraints:
    :param to_dict:

    :return:
    """

    rfba, reactions, _, _, initial_state, constraints = _inputs_processing(model=model,
                                                                           method=method,
                                                                           fraction=None,
                                                                           reactions=reactions,
                                                                           genes=None,
                                                                           regulators=None,
                                                                           objective=objective,
                                                                           initial_state=initial_state,
                                                                           constraints=constraints)

    res = {}

    for rxn in reactions:

        if isinstance(rxn, str):

            rxn_id = rxn

        else:
            rxn_id = rxn.id

        constraints[rxn_id] = (0.0, 0.0)

        sol, status = _rfba(method=rfba,
                            objective=None,
                            minimize=False,
                            initial_state=initial_state,
                            constraints=constraints,
                            status=True)

        res[rxn_id] = [sol, status]

        del constraints[rxn_id]

    if to_dict:
        return res

    return DataFrame.from_dict(data=res, orient='index', columns=['growth', 'status'])


def isingle_regulator_deletion(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                               method: str = 'srfba',
                               regulators: List[Union[str, Regulator]] = None,
                               objective: Union[str, Dict[str, Union[float, int]]] = None,
                               initial_state: Dict[str, Union[float, int]] = None,
                               constraints: Dict[str, Tuple[Union[float, int], Union[float, int]]] = None,
                               to_dict: bool = False) -> Union[Dict[str, List[Union[float, int]]], DataFrame]:
    """

    Integrated single regulator deletions

    :param model:
    :param method:
    :param regulators:
    :param objective:
    :param initial_state:
    :param constraints:
    :param to_dict:

    :return:
    """
    rfba, _, _, regulators, initial_state, constraints = _inputs_processing(model=model,
                                                                            method=method,
                                                                            fraction=None,
                                                                            reactions=None,
                                                                            genes=None,
                                                                            regulators=regulators,
                                                                            objective=objective,
                                                                            initial_state=initial_state,
                                                                            constraints=constraints)

    res = {}

    for regulator in regulators:

        if isinstance(regulator, str):

            regulator_id = regulator

        else:
            regulator_id = regulator.id

        initial_state[regulator_id] = 0.0

        sol, status = _rfba(method=rfba,
                            objective=None,
                            minimize=False,
                            initial_state=initial_state,
                            constraints=constraints,
                            status=True)

        res[regulator_id] = [sol, status]

        del initial_state[regulator_id]

    if to_dict:
        return res

    return DataFrame.from_dict(data=res, orient='index', columns=['growth', 'status'])



"""scikits ODE Solver
"""
from .ode import ODEMethod, SolverConfigurations, ODESolver
from scikits.odes.odeint import odeint
import warnings
methods = {
}


class ScikitsODESolver(ODESolver):
    """
    ODE solver method implemented on odespy package.
    """

    def __init__(self, func, method):
        self.func = func
        if method in methods.keys():
            self.method = method
        else:
            warnings.warn(f'Method {method} is unavailable.')

        self.initial_condition = None

    def set_initial_condition(self, initial_condition):
        self.initial_condition = initial_condition

    def solve(self, y0, t_span, **kwargs):
        """
        Returns the solver method from odespy package.

        :param func: function with ODE system.
        :return: an instance of odeSolver

        """
        sol = odeint(self.func, t_span, y0)
        C = [c[-1] for c in sol.y]
        return C, sol.t, sol.y



import abc

from .algebra_utils import solution_decode


def _walk(symbolic, reverse=False):
    if reverse:

        for child in symbolic.variables:
            for subtree in _walk(child, reverse):
                yield subtree

        yield symbolic

    else:

        yield symbolic

        for child in symbolic.variables:
            yield from _walk(child, reverse)


# TODO: methods stubs and type hinting
class Symbolic:

    def __init__(self, value=None, variables=None):

        if not variables:
            variables = []

        self.value = value
        self.variables = variables

        self._name = None

        self.is_boolean = False
        self.is_true = False
        self.is_false = False
        self.is_and = False
        self.is_or = False
        self.is_not = False

        self.is_relational = False
        self.is_equal = False
        self.is_not_equal = False
        self.is_inequality = False
        self.is_greater = False
        self.is_greater_equal = False
        self.is_less = False
        self.is_less_equal = False

        self.is_numeric = False
        self.is_integer = False
        self.is_float = False
        self.is_one = False
        self.is_zero = False

        self.is_symbol = False
        self.is_atom = False

        self.is_none = False

    @property
    def name(self):

        if self._name is None:

            return str(self.value)

        return self._name

    @name.setter
    def name(self, value):

        if value is None:
            return

        self._name = value

    def key(self):

        return self.to_string().replace(' ', '')

    def __repr__(self):

        if self.is_atom:

            return f"{self.__class__.__name__}({self.value})"

        else:

            return f"{self.__class__.__name__}(variables={self.variables})"

    def __str__(self):

        string_repr = ''

        if self.is_and:
            string_repr += '('
            for child in self.variables:
                string_repr += child.to_string() + ' & '
            string_repr = string_repr[:-3]
            string_repr += ')'

        elif self.is_or:
            string_repr += '('
            for child in self.variables:
                string_repr += child.to_string() + ' | '
            string_repr = string_repr[:-3]
            string_repr += ')'

        elif self.is_not:
            string_repr += '('
            for child in self.variables:
                string_repr += ' ~ ' + child.to_string()
            string_repr += ')'

        elif self.is_equal:
            string_repr += '('
            for child in self.variables:
                string_repr += child.to_string() + ' = '
            string_repr = string_repr[:-3]
            string_repr += ')'

        elif self.is_not_equal:
            string_repr += '('
            for child in self.variables:
                string_repr += child.to_string() + ' != '
            string_repr = string_repr[:-3]
            string_repr += ')'

        elif self.is_inequality:
            string_repr += '('
            for child in self.variables:
                string_repr += child.to_string() + ' > '
            string_repr = string_repr[:-3]
            string_repr += ')'

        elif self.is_greater:
            string_repr += '('
            for child in self.variables:
                string_repr += child.to_string() + ' > '
            string_repr = string_repr[:-3]
            string_repr += ')'

        elif self.is_greater_equal:
            string_repr += '('
            for child in self.variables:
                string_repr += child.to_string() + ' >= '
            string_repr = string_repr[:-3]
            string_repr += ')'

        elif self.is_less:
            string_repr += '('
            for child in self.variables:
                string_repr += child.to_string() + ' < '
            string_repr = string_repr[:-3]
            string_repr += ')'

        elif self.is_less_equal:
            string_repr += '('
            for child in self.variables:
                string_repr += child.to_string() + ' <= '
            string_repr = string_repr[:-3]
            string_repr += ')'

        else:

            string_repr += self.name

        return string_repr

    def to_string(self):
        return self.__str__()

    def __iter__(self):

        self.__variables_iter = list(_walk(self, reverse=True))
        self.__i = -1
        self.__i_limit = len(self.__variables_iter)

        return self

    def __next__(self):

        self.__i += 1

        if self.__i < self.__i_limit:
            return self.__variables_iter[self.__i]

        raise StopIteration

    def atoms(self, symbols_only=False):

        symbols = list(_walk(self))

        if symbols_only:

            return [symbol for symbol in symbols if symbol.is_symbol]

        else:

            return [symbol for symbol in symbols if symbol.is_atom]

    def _tree_evaluation(self, values, operators, default, **kwargs):

        left = None
        right = None

        if not self.variables:
            return self._evaluate(values=values, left=left, right=right, operators=operators, default=default, **kwargs)

        # noinspection PyProtectedMember
        left = self.variables[0]._tree_evaluation(values=values, operators=operators, default=default, **kwargs)

        for child in self.variables[1:]:
            # noinspection PyProtectedMember
            right = child._tree_evaluation(values=values, operators=operators, default=default, **kwargs)
            left = self._evaluate(values=values, left=left, right=right, operators=operators, default=default, **kwargs)

        # not special case
        if right is None:
            left = self._evaluate(values=values, left=left, right=right, operators=operators, default=default, **kwargs)

        return left

    def evaluate(self, values=None, operators=None, default=0, **kwargs):

        if not values:
            values = {}

        if not operators:
            operators = {}

        return self._tree_evaluation(values=values, operators=operators, default=default, **kwargs)

    @staticmethod
    @abc.abstractmethod
    def _evaluate(**kwargs):
        return

    def __call__(self, values=None, operators=None, default=0, **kwargs):

        return self.evaluate(values=values,
                             operators=operators,
                             default=default,
                             **kwargs)


class Boolean(Symbolic):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_boolean = True

    def __and__(self, *args):
        return And(variables=[self, *args])

    __rand__ = __and__

    def __or__(self, *args):
        return Or(variables=[self, *args])

    __ror__ = __or__

    def __invert__(self):
        return Not(variables=[self])

    @staticmethod
    @abc.abstractmethod
    def _evaluate(**kwargs):
        return


class BoolFalse(Boolean):

    def __init__(self, *args, **kwargs):
        if args:
            args = args[1:]

        super().__init__(*args, **kwargs)

        self.value = 0

        self.is_false = True
        self.is_atom = True

    def __bool__(self):
        return False

    @staticmethod
    def _evaluate(**kwargs):
        return 0


class BoolTrue(Boolean):

    def __init__(self, *args, **kwargs):
        if args:
            args = args[1:]

        super().__init__(*args, **kwargs)

        self.value = 1

        self.is_true = True
        self.is_atom = True

    def __bool__(self):
        return True

    @staticmethod
    def _evaluate(**kwargs):
        return 1


class And(Boolean):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_and = True

    @staticmethod
    def _evaluate(left=0, right=0, operators=None, **kwargs):

        if not operators:
            operators = {}

        operator = operators.get(And, None)

        if operator is not None:
            return operator(left, right)

        return solution_decode(int(left) & int(right))

    def __and__(self, *args):
        self.variables.extend(args)
        return self

    __rand__ = __and__


class Or(Boolean):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_or = True

    @staticmethod
    def _evaluate(left=0, right=0, operators=None, **kwargs):

        if not operators:
            operators = {}

        operator = operators.get(Or, None)

        if operator is not None:
            return operator(left, right)

        return solution_decode(int(left) | int(right))

    def __or__(self, *args):
        self.variables.extend(args)
        return self

    __ror__ = __or__


class Not(Boolean):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_not = True

    @staticmethod
    def _evaluate(left=0, operators=None, **kwargs):

        if not operators:
            operators = {}

        operator = operators.get(Not, None)

        if operator is not None:
            return operator(left)

        return solution_decode(~ int(left))


class Relational(Boolean):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_relational = True

    @staticmethod
    @abc.abstractmethod
    def _evaluate(**kwargs):
        return


class Equal(Relational):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_equal = True

    @staticmethod
    def _evaluate(left=0, right=0, operators=None, **kwargs):

        if not operators:
            operators = {}

        operator = operators.get(Equal, None)

        if operator is not None:
            return operator(left, right)

        return solution_decode(left == right)


class NotEqual(Relational):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_not_equal = True

    @staticmethod
    def _evaluate(left=0, right=0, operators=None, **kwargs):

        if not operators:
            operators = {}

        operator = operators.get(NotEqual, None)

        if operator is not None:
            return operator(left, right)

        return solution_decode(left != right)


class Inequality(Relational):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_inequality = True

    @staticmethod
    def _evaluate(left=0, right=0, operators=None, **kwargs):

        if not operators:
            operators = {}

        operator = operators.get(Inequality, None)

        if operator is not None:
            return operator(left, right)

        return solution_decode(left > right)


class Greater(Relational):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_greater = True

    @staticmethod
    def _evaluate(left=0, right=0, operators=None, **kwargs):

        if not operators:
            operators = {}

        operator = operators.get(Greater, None)

        if operator is not None:
            return operator(left, right)

        return solution_decode(left > right)


class GreaterEqual(Relational):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_greater_equal = True

    @staticmethod
    def _evaluate(left=0, right=0, operators=None, **kwargs):

        if not operators:
            operators = {}

        operator = operators.get(GreaterEqual, None)

        if operator is not None:
            return operator(left, right)

        return solution_decode(left >= right)


class Less(Relational):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_less = True

    @staticmethod
    def _evaluate(left=0, right=0, operators=None, **kwargs):

        if not operators:
            operators = {}

        operator = operators.get(Less, None)

        if operator is not None:
            return operator(left, right)

        return solution_decode(left < right)


class LessEqual(Relational):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_less_equal = True

    @staticmethod
    def _evaluate(left=0, right=0, operators=None, **kwargs):

        if not operators:
            operators = {}

        operator = operators.get(LessEqual, None)

        if operator is not None:
            return operator(left, right)

        return solution_decode(left <= right)


class AtomNumber(Symbolic):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_numeric = True
        self.is_atom = True

    def __eq__(self, other):
        return Equal(variables=[self, other])

    def __ne__(self, other):
        return NotEqual(variables=[self, other])

    def __ge__(self, other):
        return GreaterEqual(variables=[self, other])

    def __le__(self, other):
        return LessEqual(variables=[self, other])

    def __gt__(self, other):
        return Greater(variables=[self, other])

    def __lt__(self, other):
        return Less(variables=[self, other])

    @staticmethod
    @abc.abstractmethod
    def _evaluate(**kwargs):
        return


class Number(AtomNumber):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __eq__(self, other):
        return Equal(variables=[self, other])

    def __ne__(self, other):
        return NotEqual(variables=[self, other])

    def __ge__(self, other):
        return GreaterEqual(variables=[self, other])

    def __le__(self, other):
        return LessEqual(variables=[self, other])

    def __gt__(self, other):
        return Greater(variables=[self, other])

    def __lt__(self, other):
        return Less(variables=[self, other])

    @staticmethod
    @abc.abstractmethod
    def _evaluate(**kwargs):
        return


class Integer(Number):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.value = int(self.value)

        self.is_integer = True

    def _evaluate(self, **kwargs):
        return self.value


class Float(Number):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.value = float(self.value)

        self.is_float = True

    def _evaluate(self, **kwargs):
        return self.value


class Zero(Boolean, AtomNumber):

    def __init__(self, *args, **kwargs):
        if args:
            args = args[1:]

        super().__init__(*args, **kwargs)

        self.value = 0

        self.is_zero = True

    def __bool__(self):
        return False

    def _evaluate(self, **kwargs):
        return 0


class One(Boolean, AtomNumber):

    def __init__(self, *args, **kwargs):
        if args:
            args = args[1:]

        super().__init__(*args, **kwargs)

        self.value = 1

        self.is_one = True

    def __bool__(self):
        return True

    def _evaluate(self, **kwargs):
        return 1


class Symbol(AtomNumber, Boolean):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_boolean = False
        self.is_numeric = False

        self.is_symbol = True

    def _evaluate(self, values=None, default=0, **kwargs):

        if not values:
            values = {}

        value = values.get(self.name, default)

        if not isinstance(value, (int, float, bool, AtomNumber)):
            raise ValueError(f'cannot evaluate {self.name} for {value}')

        return solution_decode(value)


class NoneAtom(Symbol):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_none = True
        self.is_symbol = False

    @property
    def name(self):
        return ''

    def _evaluate(self, **kwargs):
        return 0



from collections import defaultdict
from typing import Union, TYPE_CHECKING, Tuple

from mewpy.model import Model
from .engines import JSON

if TYPE_CHECKING:
    from mewpy.model import Model, MetabolicModel, RegulatoryModel
    from .builder import Builder
    from .reader import Reader
    from .writer import Writer


class Director:

    def __init__(self, *builders: Union['Builder', 'Reader', 'Writer']):
        """

        Director is responsible for managing builders. It gives instructions to the builders on how to read or write
        a multi-type model properly

        To understand how reading and writing are proceeded, see the read and write methods

        :type builders: Union['Builder', 'Reader', 'Writer']

        :param builders: builders for reading models, files or IOs. builders for writing files using mewpy models
        """
        self._builders = builders

    @property
    def builders(self) -> Tuple[Union['Builder', 'Reader', 'Writer']]:

        """
        Returns the builders associated with this director
        """

        return self._builders

    def read(self) -> Union['Model', 'RegulatoryModel', 'MetabolicModel']:

        """
        Reading a mewpy model, namely metabolic, regulatory or both encoded into one or more file types.
        Reading is performed step-wise according to the builders order.

        """

        types = set()
        variables = defaultdict(set)

        for builder in self.builders:

            engine = builder.engine

            # First, opening the resource in read mode
            engine.open(mode='r')

            # Second, retrieving the model type
            types.add(engine.model_type)

            # Third, parsing the resource

            # parsing is responsible for two main operations:
            #   - reading a file or model with the adequate engine (pandas, python-lib-sbml, cobra, etc)
            #   and parse all important items to a data transfer object, the dto.

            #   - while reading the file or model, all variables and their types across the various files and models are
            #   collected and centralized under the engine.variables property
            engine.parse()

            # merging all variables and their types across the various files and models
            for variable, var_types in engine.variables.items():
                variables[variable].update(var_types)

            # Fourth, the resource opened by the engine are closed and can be used next
            engine.close()

        # Fifth, the model is created, as we now know the model types.
        model = Model.from_types(types=types, identifier='model')

        for builder in self.builders:

            engine = builder.engine

            if isinstance(engine, JSON):

                # json is a special case because of serialization

                model = engine.read()
                engine.clean()
                return model

            # Sixth, model building takes place.

            # The model is incremented with further containers and attributes, namely the stuff that each file/model
            # has to offer

            # In detail, when the read method of an engine receives a model object, it does not create a new one.
            # Instead, it adds new variables to the already available attributes and containers

            # In detail, when the read method of a engine receives a dictionary of a set of variables, it will create
            # the variables according to the variables types listed among all files.
            # That is, if var_01 has target, regulator and metabolite types registered across the various files/models,
            # the read method will create a TargetRegulatorMetabolite variable under a single instance
            # (a multi-type variable). This variable contains all attributes of a Target, Regulator and Metabolite
            engine.read(model=model, variables=variables)

            # Seventh, cleaning the data transfer object In detail, when the open and parse methods of an engine are
            # used, the data transfer object is populated with records. Then, it is no longer required for this
            # object to live in memory
            engine.clean()

        model.clean_history()

        return model

    def write(self):

        """
        Writing a mewpy model, namely metabolic, regulatory or both to one or more file types.
        Writing is performed step-wise according to the builders order.

        """

        for builder in self.builders:

            engine = builder.engine

            # First, opening the resource in write mode
            engine.open(mode='w')

            # Second, writing the model to the file
            engine.write()

            # Third, closing the resource
            engine.close()

            # Seventh, cleaning the data transfer object In detail, when the open and write methods of an engine are
            # used, the data transfer object is populated with records. Then, it is no longer required for this
            # object to live in memory
            engine.clean()

    def warn(self):

        """
        Launch warnings stored in the builders.

        """

        # warnings and even small errors are all collected with partial pattern and then launched, as printing to the
        # console takes time

        for builder in self.builders:

            for warning in builder.warnings:
                warning()



from .symbolic import (BoolTrue,
                       BoolFalse,
                       And,
                       Or,
                       Not,
                       Equal,
                       NotEqual,
                       Inequality,
                       Greater,
                       GreaterEqual,
                       Less,
                       LessEqual,
                       Integer,
                       Float,
                       Zero,
                       One,
                       Symbol,
                       NoneAtom)

# Boolean escape chars to be replaced
BOOLEAN_ESCAPE_CHARS = {'-': '_dash_',
                        ',': '_comma_',
                        ';': '_semicolon_',
                        '[': '_lbracket_',
                        ']': '_rbracket_',
                        ':': '_colon_',
                        '+': '_plus_',
                        '/': '_slash_',
                        '\\': '_backslash_',
                        '$': '_dollar_',
                        '%': '_percentage_',
                        '"': '_quotes_',
                        '(e)': '_e_',
                        '(c)': '_c_',
                        '(p)': '_p_'}

RELATIONAL_ESCAPE_CHARS = BOOLEAN_ESCAPE_CHARS

# Boolean string operator symbols
AND = '&'
OR = '|'
NOT = '~'

BOOLEAN_OPERATORS = {'not': NOT,
                     '!': NOT,
                     'and': AND,
                     'or': OR,
                     AND: AND,
                     OR: OR,
                     NOT: NOT}

GLOBAL_BOOLEAN_OPERATORS = {AND, OR, NOT}

# Boolean states
TRUE = '1'
FALSE = '0'

BOOLEAN_STATES = {'on': TRUE,
                  'off': FALSE,
                  'true': TRUE,
                  'false': FALSE,
                  TRUE: TRUE,
                  FALSE: FALSE}

GLOBAL_BOOLEAN_STATES = {TRUE, FALSE}

# Strict relational operators
GREATER = '>'
LESS = '<'
EQUAL = '='
NOT_EQUAL = '!='

RELATIONAL_OPERATORS = {'greater than': GREATER,
                        'less than': LESS,
                        'greater': GREATER,
                        'less': LESS,
                        '==': EQUAL,
                        'equal': EQUAL,
                        'not equal': NOT_EQUAL,
                        GREATER: GREATER,
                        LESS: LESS,
                        EQUAL: EQUAL,
                        NOT_EQUAL: NOT_EQUAL}

GLOBAL_RELATIONAL_OPERATORS = {GREATER, LESS, EQUAL, NOT_EQUAL}

# Relational operators
GREATER_EQUAL = '>='
LESS_EQUAL = '=>'

RELATIONAL_EQUAL_OPERATORS = {'greater than or equal': GREATER_EQUAL,
                              'less than or equal': LESS_EQUAL,
                              'greater or equal': GREATER_EQUAL,
                              'less or equal': LESS_EQUAL,
                              GREATER_EQUAL: GREATER_EQUAL,
                              LESS_EQUAL: LESS_EQUAL}

GLOBAL_RELATIONAL_EQUAL_OPERATORS = {GREATER_EQUAL, LESS_EQUAL}

RELATIONAL_STATES = BOOLEAN_STATES

GLOBAL_RELATIONAL_STATES = GLOBAL_BOOLEAN_STATES

GLOBAL_MEWPY_OPERATORS = {'BoolFalse': BoolFalse,
                          'BoolTrue': BoolTrue,
                          'And': And,
                          'Or': Or,
                          'Not': Not,
                          'Equal': Equal,
                          'NotEqual': NotEqual,
                          'Inequality': Inequality,
                          'Greater': Greater,
                          'GreaterEqual': GreaterEqual,
                          'Less': Less,
                          'LessEqual': LessEqual,
                          'Integer': Integer,
                          'Float': Float,
                          'Zero': Zero,
                          'One': One,
                          'Symbol': Symbol,
                          'NoneAtom': NoneAtom}



""" Adapted from Daniel Machado's REFRAMED"""
from enum import Enum
import re


class Status(Enum):
    """ Enumeration of possible solution status. """
    OPTIMAL = 'Optimal'
    UNKNOWN = 'Unknown'
    SUBOPTIMAL = 'Suboptimal'
    UNBOUNDED = 'Unbounded'
    INFEASIBLE = 'Infeasible'
    INF_OR_UNB = 'Infeasible or Unbounded'


class Solution(object):
    """ Stores the results of an optimization.

    Instantiate without arguments to create an empty Solution representing a failed optimization.
    """

    def __init__(self, status=Status.UNKNOWN, message=None, fobj=None, values=None,
                 shadow_prices=None, reduced_costs=None):
        self.status = status
        self.message = message
        self.fobj = fobj
        self.values = values
        self.shadow_prices = shadow_prices
        self.reduced_costs = reduced_costs

    def __str__(self):
        return f"Objective: {self.fobj}\nStatus: {self.status.value}\n"

    def __repr__(self):
        return str(self)

    def to_dataframe(self):
        """ Convert reaction fluxes to *pandas.DataFrame*

        Returns:
            pandas.DataFrame: flux values
        """
        try:
            import pandas as pd
        except ImportError:
            raise RuntimeError("Pandas is not installed.")

        return pd.DataFrame(self.values.values(), columns=["value"], index=self.values.keys())



import math
import networkx as nx
import numpy as np
from ..simulation import get_simulator
from ..simulation.simulation import Simulator

METABOLITE = 'metabolite'
REACTION = 'reaction'
REV = 'rev'
IRREV = 'irrev'


def create_metabolic_graph(model, directed=True, carbon=True, reactions=None, remove=[], edges_labels=False, biomass=False, metabolites=False):
    """ Creates a metabolic graph

    :param model: A model or a model containter
    :param (bool) directed: Defines if the graph to be directed or undirected. Defaults to True.
    :param (bool) carbon: Only included edges for metabolites with a carbon atom. Defaults to True.
    :param (list) reactions: List of reactions to be included in the graph. Defaults to None, in which\
        all reactions are included.
    :param list remove: list os metabolites not to be included. May be used to remove cofactores such as ATP/ADP, \
        NAD(P)(H), and acetyl-CoA/CoA.
    :param (bool) edges_labels: Adds a reversabily label to edges. Defaults to False.
    :returns: A networkx graph of the metabolic network.
    """

    if not isinstance(model, Simulator):
        container = get_simulator(model)
    else:
        container = model

    if directed:
        G = nx.DiGraph()
    else:
        G = nx.Graph()
    if not reactions:
        reactions = container.reactions

    reactions = list(set(reactions) - set(remove))

    for r in reactions:
        G.add_node(r, label=r, node_class=REACTION, node_id=r)

    for r in reactions:
        the_metabolites = container.get_reaction_metabolites(r)
        for m in the_metabolites:
            if m in remove:
                continue
            if carbon and 'C' not in container.metabolite_elements(m).keys():
                continue
            if m not in G.nodes:
                G.add_node(m, label=m, node_class=METABOLITE, node_id=m)
            # evaluating if the metabolite has been defined as a reactant or product
            if the_metabolites[m] < 0:
                (tail, head) = (m, r)
            elif the_metabolites[m] > 0:
                (tail, head) = (r, m)

            # adding an arc between a metabolite and a reactions
            G.add_edge(tail, head)
            label = IRREV
            lb, _ = container.get_reaction_bounds(r)

            if lb < 0:
                G.add_edge(head, tail)
                label = REV

            if edges_labels:
                G[tail][head]['label'] = label

            G[tail][head]['reversible'] = lb < 0

    if not metabolites:
        met_nodes = [x for x, v in dict(G.nodes(data="node_class")).items() if v == METABOLITE]
        for m in met_nodes:
            in_ = G.in_edges(m, data=True)
            out_ = G.out_edges(m, data=True)
            for s, _, r1 in in_:
                for _, t, r2 in out_:
                    try:
                        rev = r1['reversible'] and r2['reversible']
                    except:
                        rev = False
                    G.add_edge(s, t)
            G.remove_node(m)

    return G


def filter_by_degree(G, max_degree, inplace=True):
    s = list(sorted(G.degree, key=lambda item: item[1], reverse=True))
    stop = False
    while not stop:
        # find the metabolite with highest degree
        print(s[:5])
        position = 0
        found = False
        k = None
        v = None
        while not found and position < len(s):
            k, v = s[position]
            if G.nodes[k]['node_class'] == METABOLITE:
                found = True
            else:
                position += 1
        if k and v > max_degree:
            G.remove_node(k)
            print('removed ', k)
            s = list(sorted(G.degree, key=lambda item: item[1], reverse=True))
        else:
            stop = True
    return G


def shortest_distance(model, reaction, reactions=None, remove=[]):
    """ Returns the unweighted shortest path distance from a list of reactions to a reaction.
    Distances are the number of required reactions. If there is no pathway between the reactions the distance is inf·

    :param model: A model or a Simulator instance.
    :param str reaction: target reaction.
    :param list reactions: List os source reactions. Defaults to None, in which case all model reactions are considered.
    :param list remove: List os metabolites not to be included. May be used to remove path that include \
        cofactores such as ATP/ADP, NAD(P)(H), and acetyl-CoA/CoA.
    :returns: A dictionary of distances.
    """
    if not isinstance(model, Simulator):
        container = get_simulator(model)
    else:
        container = model

    rxns = reactions if reactions else container.reactions
    if reaction not in rxns:
        rxns.append(reaction)

    G = create_metabolic_graph(container, reactions=rxns, remove=remove)
    sp = dict(nx.single_target_shortest_path_length(G, reaction))

    distances = {}
    for rxn in rxns:
        if rxn in sp:
            distances[rxn] = sp[rxn] // 2
        else:
            distances[rxn] = np.inf
    return distances


def probabilistic_reaction_targets(model, product, targets, factor=10):
    """Builds a new target list reflecting the shortest path distances from all original
    as a probability,ie, reactions closer to the product are repeated more often in the new target list.
    Moreover, reactions from which there is no path (pathway or cofactors usage) to the product are removed.

    :param model: A model or a Simulator instance.
    :param str product: Product to be optimized.
    :param str targets: EA target reactions.
    :param int factor: Maximum number of repetitions, also the distance after which all reactions are\
        considered with equal probability. Defaults to 10.
    :returns: A probabilistic target list.
    """
    distances = shortest_distance(model, product, targets)
    prob_targets = []
    for t in targets:
        if distances[t] == np.inf or distances[t] == 0:
            continue
        else:
            coef = math.ceil(1 / distances[t] * factor)
        x = [t] * coef
        prob_targets.extend(x)
    return prob_targets


def probabilistic_gene_targets(model, product, targets, factor=10):
    """Builds a new target list reflecting the shortest path distances from all original
    as a probability,ie, genes on GPRs of reactions closer to the product are repeated more
    often in the new target list.

    :param model: A model or a Simulator instance.
    :param str product: Product to be optimized.
    :param str targets: EA target genes.
    :param int factor: Maximum number of repetitions. Defaults to 10.
    :returns: A probabilistic target list.
    """
    if not isinstance(model, Simulator):
        container = get_simulator(model)
    else:
        container = model

    # Reaction targets
    if not targets:
        genes = container.genes
    else:
        genes = targets

    rxns = container.get_reactions_for_genes(genes)
    rxn_distances = shortest_distance(model, product, rxns)

    # genes distances are the maximum of all reaction
    # distances that they catalyse.

    prob_targets = []

    for gene in genes:
        rxs = container.get_reactions_for_genes([gene])
        dd = [d for r, d in rxn_distances.items() if r in rxs]
        d = max(dd)
        if d == np.inf:
            coef = 1
        elif d == 0:
            continue
        else:
            coef = math.ceil(1 / d * factor)
        x = [gene] * coef
        prob_targets.extend(x)

    return prob_targets


def probabilistic_protein_targets(model, product, targets, factor=10):
    """Builds a new target list reflecting the shortest path distances from all original
    as a probability,ie, proteins used in reactions closer to the product are repeated
    more often in the new target list.

    :param model: A model or a Simulator instance.
    :param str product: Product to be optimized.
    :param str targets: EA target genes.
    :param int factor: Maximum number of repetitions. Defaults to 10.
    :returns: A probabilistic target list.
    """
    raise NotImplementedError



from typing import Any, TYPE_CHECKING, Union, Generator, Dict, List, Tuple, Set

from mewpy.model.model import Model
from mewpy.mew.lp import Notification
from mewpy.util.history import recorder
from mewpy.util.serialization import serialize
from mewpy.util.utilities import iterable, generator

if TYPE_CHECKING:
    from mewpy.mew.variables import Interaction, Regulator, Target, Metabolite, Reaction


# TODO: methods stubs
class RegulatoryModel(Model, model_type='regulatory', register=True, constructor=True, checker=True):

    def __init__(self,
                 identifier: Any,
                 compartments: Dict[str, str] = None,
                 interactions: Dict[str, 'Interaction'] = None,
                 regulators: Dict[str, 'Regulator'] = None,
                 targets: Dict[str, 'Target'] = None,
                 **kwargs):

        """
        A regulatory model contains essentially interactions, regulators and targets.
        The regulatory model can be loaded with compartments, although these can be inferred from the available
        regulators.

        Other information is retrieved from these attributes, namely the regulatory events happening in the model

        The regulatory model, as with other models, provides a clean interface for manipulation with the add, remove and
        update methods.

        :param identifier: identifier, e.g. lac_operon_regulation
        :param compartments: a dictionary with additional compartments not encoded in the regulators
        :param interactions: a dictionary with Interaction objects. See variables.Interaction for more info
        :param regulators: a dictionary with Regulator objects. See variables.Regulator for more info
        :param targets: a dictionary with Target objects. See variables.Target for more info
        """

        # compartments attribute can be shared across the children, thus name mangling
        self.__compartments = {}
        self._interactions = {}
        self._regulators = {}
        self._targets = {}

        super().__init__(identifier,
                         **kwargs)

        # the setters will handle adding and removing variables to the correct containers
        self.compartments = compartments
        self.interactions = interactions
        self.regulators = regulators
        self.targets = targets

    # -----------------------------------------------------------------------------
    # Model type manager
    # -----------------------------------------------------------------------------

    @serialize('types', None)
    @property
    def types(self):
        # noinspection PyUnresolvedReferences
        _types = {RegulatoryModel.model_type}

        _types.update(super(RegulatoryModel, self).types)

        return _types

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------

    @serialize('interactions', 'interactions', '_interactions')
    @property
    def interactions(self) -> Dict[str, 'Interaction']:
        return self._interactions.copy()

    @serialize('regulators', 'regulators', '_regulators')
    @property
    def regulators(self) -> Dict[str, 'Regulator']:
        return self._regulators.copy()

    @serialize('targets', 'targets', '_targets')
    @property
    def targets(self) -> Dict[str, 'Target']:
        return self._targets.copy()

    @property
    def compartments(self) -> Dict[str, str]:

        compartments = {regulator.compartment: self.__compartments.get(regulator.compartment, '')
                        for regulator in self.yield_regulators()
                        if regulator.is_metabolite() and regulator.compartment is not None}

        compartments.update(self.__compartments)

        compartments.update(super(RegulatoryModel, self).compartments)

        return compartments

    # -----------------------------------------------------------------------------
    # Static attributes setters
    # -----------------------------------------------------------------------------

    @compartments.setter
    @recorder
    def compartments(self, value: Dict[str, str]):

        if not value:
            value = {}

        self.__compartments.update(value)

    @interactions.setter
    @recorder
    def interactions(self, value: Dict[str, 'Interaction']):

        if not value:
            value = {}

        self.remove(list(self.yield_interactions()), 'interaction', history=False)
        self.add(list(value.values()), 'interaction', history=False)

    @regulators.setter
    @recorder
    def regulators(self, value: Dict[str, 'Regulator']):

        if not value:
            value = {}

        self.remove(list(self.yield_regulators()), 'regulator', history=False)
        self.add(list(value.values()), 'regulator', history=False)

    @targets.setter
    @recorder
    def targets(self, value: Dict[str, 'Target']):

        if not value:
            value = {}

        self.remove(list(self.yield_targets()), 'target', history=False)
        self.add(list(value.values()), 'target', history=False)

    # -----------------------------------------------------------------------------
    # Dynamic attributes
    # -----------------------------------------------------------------------------

    @property
    def environmental_stimuli(self) -> Dict[str, 'Regulator']:

        return {reg_id: regulator for reg_id, regulator in self.regulators.items()
                if regulator.environmental_stimulus}

    # -----------------------------------------------------------------------------
    # Generators
    # -----------------------------------------------------------------------------

    def yield_environmental_stimuli(self) -> Generator['Regulator', None, None]:

        return generator(self.environmental_stimuli)

    def yield_interactions(self) -> Generator['Interaction', None, None]:

        return generator(self._interactions)

    def yield_regulators(self) -> Generator[Union['Regulator', 'Metabolite', 'Reaction'], None, None]:

        return generator(self._regulators)

    def yield_targets(self) -> Generator['Target', None, None]:

        return generator(self._targets)

    # -----------------------------------------------------------------------------
    # Operations/Manipulations
    # -----------------------------------------------------------------------------

    def get(self, identifier: Any, default=None) -> Union['Interaction', 'Regulator', 'Target']:

        if identifier in self._targets:
            return self._targets[identifier]

        elif identifier in self._regulators:
            return self._regulators[identifier]

        elif identifier in self._interactions:
            return self._interactions[identifier]

        else:
            return super(RegulatoryModel, self).get(identifier=identifier, default=default)

    def add(self,
            variables: Union[List[Union['Interaction', 'Regulator', 'Target']],
                             Tuple[Union['Interaction', 'Regulator', 'Target']],
                             Set[Union['Interaction', 'Regulator', 'Target']]],
            *types: str,
            comprehensive: bool = True,
            history=True):

        variables = iterable(variables)

        if not types:
            types = [var.types for var in variables]

        elif len(types) == len(variables):
            types = [{_type} if isinstance(_type, str) else set(_type) for _type in types]

        elif len(types) != len(variables):
            types = [set(types) for var in variables]

        interactions = []
        new_variables = []
        new_types = []

        for var_types, var in zip(types, variables):

            if 'target' in var_types:
                self._add_target(var)
                var_types.remove('target')

            elif 'regulator' in var_types:
                self._add_regulator(var)
                var_types.remove('regulator')

            elif 'interaction' in var_types:
                self._add_interaction(var, comprehensive=comprehensive)
                interactions.append(var)
                var_types.remove('interaction')

            if var_types:
                new_types.append(var_types)
                new_variables.append(var)

        if interactions:
            notification = Notification(content=interactions,
                                        content_type='interactions',
                                        action='add')

            self.notify(notification)

        if history:
            self.history.queue_command(undo_func=self.remove,
                                       undo_kwargs={'variables': variables,
                                                    'remove_orphans': True,
                                                    'history': False},
                                       func=self.add,
                                       kwargs={'variables': variables,
                                               'comprehensive': comprehensive,
                                               'history': history})

        super(RegulatoryModel, self).add(new_variables,
                                         *new_types,
                                         comprehensive=comprehensive,
                                         history=False)

    def remove(self,
               variables: Union[List[Union['Interaction', 'Regulator', 'Target']],
                                Tuple[Union['Interaction', 'Regulator', 'Target']],
                                Set[Union['Interaction', 'Regulator', 'Target']]],
               *types: str,
               remove_orphans: bool = False,
               history=True):

        variables = iterable(variables)

        if not types:
            types = [var.types for var in variables]

        elif len(types) == len(variables):
            types = [{_type} if isinstance(_type, str) else set(_type) for _type in types]

        elif len(types) != len(variables):
            types = [set(types) for var in variables]

        interactions = []
        new_variables = []
        new_types = []

        for var_types, var in zip(types, variables):

            if 'target' in var_types:
                self._remove_target(var)
                var_types.remove('target')

            elif 'regulator' in var_types:
                self._remove_regulator(var)
                var_types.remove('regulator')

            elif 'interaction' in var_types:
                self._remove_interaction(var, remove_orphans=remove_orphans)
                interactions.append(var)
                var_types.remove('interaction')

            if var_types:
                new_types.append(var_types)
                new_variables.append(var)

        if interactions:

            notification = Notification(content=interactions,
                                        content_type='interactions',
                                        action='remove')

            self.notify(notification)

            if remove_orphans:
                self._remove_regulatory_orphans(interactions)

        if history:
            self.history.queue_command(undo_func=self.add,
                                       undo_kwargs={'variables': variables,
                                                    'comprehensive': True,
                                                    'history': False},
                                       func=self.remove,
                                       kwargs={'variables': variables,
                                               'remove_orphans': remove_orphans,
                                               'history': history})

        super(RegulatoryModel, self).remove(new_variables,
                                            *new_types,
                                            remove_orphans=remove_orphans,
                                            history=False)

    def update(self,
               compartments: Dict[str, str] = None,
               variables: Union[List[Union['Interaction', 'Regulator', 'Target']],
                                Tuple[Union['Interaction', 'Regulator', 'Target']],
                                Set[Union['Interaction', 'Regulator', 'Target']]] = None,
               **kwargs):

        if compartments is not None:
            self.compartments = compartments

        if variables is not None:
            self.add(variables=variables)

        super(RegulatoryModel, self).update(**kwargs)

    def _add_interaction(self, interaction, comprehensive=True):

        # adding a interaction is regularly a in-depth append method, as both regulators and target associated with
        # the interaction are also regularly added to the model. Although, this behaviour can be avoided passing
        # comprehensive=True

        if interaction.id not in self._interactions:

            if comprehensive:

                for regulator in interaction.yield_regulators():
                    self._add_regulator(regulator)

                if interaction.target is not None:
                    self._add_target(interaction.target)

            self._add_variable_to_container(interaction, self._interactions)

    def _add_regulator(self, regulator):

        if regulator.id not in self._regulators:
            self._add_variable_to_container(regulator, self._regulators)

    def _add_target(self, target):

        if target.id not in self._targets:
            self._add_variable_to_container(target, self._targets)

    def _remove_interaction(self, interaction, remove_orphans=False):

        if interaction.id in self._interactions:
            self._remove_variable_from_container(interaction, self._interactions)

    def _remove_regulator(self, regulator):

        if regulator.id in self._regulators:
            self._remove_variable_from_container(regulator, self._regulators)

    def _remove_target(self, target):

        if target.id in self._targets:
            self._remove_variable_from_container(target, self._targets)

    def _remove_regulatory_orphans(self, interactions):

        orphan_regs = self._get_orphans(to_remove=interactions,
                                        first_container='regulators',
                                        second_container='interactions')

        orphan_targets = self._get_orphans(to_remove=interactions,
                                           first_container='target',
                                           second_container='interaction')

        if orphan_regs:

            for regulator in orphan_regs:
                self._remove_regulator(regulator)

        if orphan_targets:

            for target in orphan_targets:
                self._remove_target(target)



from random import Random
from time import time

import inspyred
from .settings import get_population_size, KO, PARAMETERS, OU
from .problem import InspyredProblem
from .observers import results_observer, VisualizerObserver
from .terminator import generation_termination
from ..ea import AbstractEA, Solution
from ...util.constants import EAConstants
from ...util.process import get_evaluator, cpu_count
from .operators import OPERATORS
SOEA = {
    'GA': inspyred.ec.EvolutionaryComputation,
    'SA': inspyred.ec.SA
}


class EA(AbstractEA):
    """
    EA running helper

    :param problem: the optimization problem.
    :param initial_population: (list) the EA initial population.
    :param max_generations: (int) the number of iterations of the EA (stopping criteria).
    """

    def __init__(self, problem, initial_population=[], max_generations=EAConstants.MAX_GENERATIONS, mp=True,
                 visualizer=False, algorithm=None, **kwargs):

        super(EA, self).__init__(problem, initial_population=initial_population,
                                 max_generations=max_generations, mp=mp, visualizer=visualizer, **kwargs)

        self.algorithm_name = algorithm
        self.ea_problem = InspyredProblem(self.problem)

        # operators
        if self.problem.operators:
            self.variators = [OPERATORS[x] for x in self.problem.operators.keys()]
        elif self.problem.strategy == 'OU':
            self.variators = OU['variators']
        elif self.problem.strategy == 'KO':
            self.variators = KO['variators']
        else:
            raise ValueError("Unknow strategy")

        self.population_size = kwargs.get('population_size', get_population_size())

        # parameters
        self.args = PARAMETERS.copy()
        if self.problem.operators_param:
            self.args.update(self.problem.operators_param)
        self.args['num_selected'] = self.population_size
        self.args['max_generations'] = max_generations,
        self.args['candidate_min_size'] = self.problem.candidate_min_size
        self.args['candidate_max_size'] = self.problem.candidate_max_size
        if self.problem.number_of_objectives != 1:
            self.args.pop('tournament_size')
        self.seeds = [self.problem.encode(s) for s in initial_population]

    def get_population_size(self):
        return self.population_size

    def _run_so(self):
        """ Runs a single objective EA optimization
        """
        prng = Random()
        prng.seed(time())

        if self.mp:
            self.evaluator = get_evaluator(self.ea_problem, n_mp=cpu_count())
        else:
            self.evaluator = self.ea_problem.evaluator

        if self.algorithm_name == 'SA':
            ea = inspyred.ec.SA(prng)
            print("Running SA")
        else:
            ea = inspyred.ec.EvolutionaryComputation(prng)
            print("Running GA")
        ea.selector = inspyred.ec.selectors.tournament_selection

        ea.variator = self.variators
        ea.observer = results_observer
        ea.replacer = inspyred.ec.replacers.truncation_replacement
        ea.terminator = generation_termination
        self.algorithm = ea
        final_pop = ea.evolve(generator=self.problem.generator,
                              evaluator=self.evaluator,
                              pop_size=self.population_size,
                              seeds=self.seeds,
                              maximize=self.problem.is_maximization,
                              bounder=self.problem.bounder,
                              **self.args
                              )
        self.final_population = final_pop
        return final_pop

    def _run_mo(self):
        """ Runs a multi objective EA (NSGAII) optimization
        """
        prng = Random()
        prng.seed(time())

        if self.mp:
            self.evaluator = get_evaluator(self.ea_problem, n_mp=cpu_count())
        else:
            self.evaluator = self.ea_problem.evaluator

        ea = inspyred.ec.emo.NSGA2(prng)
        print("Running NSGAII")
        ea.variator = self.variators
        ea.terminator = generation_termination
        if self.visualizer:
            axis_labels = [f.short_str() for f in self.problem.fevaluation]
            observer = VisualizerObserver(axis_labels=axis_labels)
            ea.observer = observer.update
        else:
            ea.observer = results_observer
        self.algorithm = ea
        final_pop = ea.evolve(generator=self.problem.generator,
                              evaluator=self.evaluator,
                              pop_size=self.population_size,
                              seeds=self.seeds,
                              maximize=self.problem.is_maximization,
                              bounder=self.problem.bounder,
                              **self.args
                              )

        self.final_population = final_pop
        return final_pop

    def _convertPopulation(self, population):
        """Converts a population represented in Inpyred format to
        MEWpy solution format.

        :param list population: A list of solutions.

        :returns: A MEWpy list of solutions.
        """
        p = []
        for i in range(len(population)):
            if self.problem.number_of_objectives == 1:
                obj = [population[i].fitness]
            else:
                obj = population[i].fitness.values
            val = population[i].candidate
            values = self.problem.decode(val)
            const = self.problem.solution_to_constraints(values)
            solution = Solution(values, obj, const)
            p.append(solution)
        return p

    def _get_current_population(self):
        """Dumps the population for gracefull exit."""
        pop = self.algorithm.population
        cv = self._convertPopulation(pop)
        return cv



from typing import Any, TYPE_CHECKING, Dict, Union, Generator, Tuple

# TODO: this module depends on pandas dataframes. Should it be set as package requirement?
try:
    # noinspection PyPackageRequirements
    from pandas import concat, DataFrame

except ImportError:

    concat = False
    DataFrame = False

from mewpy.mew.algebra import Expression, parse_expression
from mewpy.mew.lp import Notification
from mewpy.util.utilities import generator
from mewpy.util.serialization import serialize
from mewpy.util.history import recorder
from mewpy.io.engines.engines_utils import expression_warning
from .variable import Variable, variables_from_symbolic


# Preventing circular dependencies that only happen due to type checking
if TYPE_CHECKING:
    from .regulator import Regulator
    from .target import Target
    from mewpy.model import Model, MetabolicModel, RegulatoryModel


# TODO: methods stubs
class Interaction(Variable, variable_type='interaction', register=True, constructor=True, checker=True):

    def __init__(self,
                 identifier: Any,
                 target: 'Target' = None,
                 regulatory_events: Dict[Union[float, int], Expression] = None,
                 **kwargs):

        """
        A regulatory interaction is regularly associated with a target and the regulatory
        events modelling the coefficients of this target variable.

        The regulatory events determine the multiple coefficients that the target variable can take
        depending on the logic and conditions encoded into an expression object.

        The expression must hold the multiple variables/regulators that participate in the regulatory event/expression.

        The set of regulators is inferred from the regulatory events.
        Also, the regulatory events are used to generate a regulatory table,
        namely the possible coefficients of the target for the active (or not) coefficients of the regulators.

        :param identifier: the interaction identifier
        :param target: the target variable that is regulated by the regulators/expressions logic
        :param regulatory_events: A dictionary comprehending coefficient-expression pairs.
        The expression when evaluated to true results into the corresponding coefficient
        that must be associated with the target coefficients

        """

        if not target:
            target = None

        if not regulatory_events:
            regulatory_events = {0.0: Expression()}

        self._regulatory_events = {}
        self._target = None

        super().__init__(identifier,
                         **kwargs)

        # it handles addition of a target. It fulfills the correct containers/attributes
        self.add_target(target, history=False)

        # it handles setting a regulatory event. It fulfills the correct containers/attributes for the regulators
        # associated with the expression og a given event
        for coefficient, expression in regulatory_events.items():
            self.add_regulatory_event(coefficient=coefficient, expression=expression, history=False)

        # There is an alternative constructor for building an interaction with a stringify-like expression rule.
        # Parsing takes the hard job to infer the correct relation between regulators

    # -----------------------------------------------------------------------------
    # Variable type manager
    # -----------------------------------------------------------------------------

    @property
    def types(self):

        # noinspection PyUnresolvedReferences
        _types = {Interaction.variable_type}

        _types.update(super(Interaction, self).types)

        return _types

    # -----------------------------------------------------------------------------
    # Built-in
    # -----------------------------------------------------------------------------

    def __str__(self):

        if self.target:
            string_repr = f'{self.target.id}: '

        else:
            string_repr = ''

        for expression in self.yield_expressions():
            string_repr += expression.to_string()

        return string_repr

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------

    @serialize('regulatory_events', 'regulatory_events', '_regulatory_events')
    @property
    def regulatory_events(self) -> Dict[Union[float, int], Expression]:
        return self._regulatory_events.copy()

    @serialize('target', 'target', '_target')
    @property
    def target(self) -> 'Target':
        return self._target

    # -----------------------------------------------------------------------------
    # Static attributes setters
    # -----------------------------------------------------------------------------

    @regulatory_events.setter
    @recorder
    def regulatory_events(self, value: Dict[Union[float, int], Expression]):

        for coefficient, expression in self.regulatory_events.items():
            self.remove_regulatory_event(coefficient=coefficient, expression=expression,
                                         remove_orphans=True, history=False)

        for coefficient, expression in value.items():
            self.add_regulatory_event(coefficient=coefficient, expression=expression, history=False)

    @target.setter
    @recorder
    def target(self, value: 'Target'):

        self.remove_target(remove_from_model=True, history=False)
        self.add_target(value, history=False)

    # -----------------------------------------------------------------------------
    # Dynamic attributes
    # -----------------------------------------------------------------------------

    @property
    def regulators(self) -> Dict[str, 'Regulator']:

        return {reg_id: regulator
                for expression in self.yield_expressions()
                for reg_id, regulator in expression.variables.items()}

    def _compute_regulatory_table(self, active_states=True):

        if concat is False:
            raise RuntimeError('pandas must be installed to compute regulatory tables')

        truth_tables = []

        for coefficient, expression in self._regulatory_events.items():
            df = expression.truth_table(active_states=active_states, coefficient=coefficient)

            df.index = [self.target.id if self.target else 'result'] * df.shape[0]

            truth_tables.append(df)

        # noinspection PyCallingNonCallable
        regulatory_events = concat(truth_tables)

        regulatory_events = regulatory_events[['result'] + [col for col in regulatory_events.columns
                                                            if col != 'result']]

        return regulatory_events

    @property
    def regulatory_table(self) -> DataFrame:

        return self._compute_regulatory_table(active_states=True)

    def full_regulatory_table(self) -> DataFrame:

        return self._compute_regulatory_table(active_states=False)

    # -----------------------------------------------------------------------------
    # Generators
    # -----------------------------------------------------------------------------

    def yield_regulatory_events(self) -> Generator[Tuple[Union[float, int], Expression], None, None]:

        return ((coefficient, expression) for coefficient, expression in self._regulatory_events.items())

    def yield_expressions(self) -> Generator[Expression, None, None]:

        return generator(self._regulatory_events)

    def yield_regulators(self) -> Generator['Regulator', None, None]:

        return generator(self.regulators)

    # -----------------------------------------------------------------------------
    # Polymorphic constructors
    # -----------------------------------------------------------------------------

    @classmethod
    def from_string(cls,
                    identifier: Any,
                    stringify_rule: str,
                    target: 'Target',
                    coefficient: Union[float, int] = 1.0,
                    model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None,
                    **kwargs) -> 'Interaction':

        try:

            symbolic = parse_expression(stringify_rule)

            regulators = variables_from_symbolic(symbolic=symbolic, types=('regulator',), model=model)

            expression = Expression(symbolic=symbolic, variables=regulators)

        except SyntaxError as exc:

            expression_warning(f'{stringify_rule} cannot be parsed')

            raise exc

        instance = cls(identifier=identifier,
                       target=target,
                       regulatory_events={coefficient: expression},
                       model=model,
                       **kwargs)

        return instance

    @classmethod
    def from_expression(cls,
                        identifier: Any,
                        expression: Expression,
                        target: 'Target',
                        coefficient: Union[float, int] = 1.0,
                        model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None,
                        **kwargs) -> 'Interaction':

        if not isinstance(expression, Expression):
            raise TypeError(f'expression must be an {Expression} object')

        instance = cls(identifier=identifier,
                       target=target,
                       regulatory_events={coefficient: expression},
                       model=model,
                       **kwargs)

        return instance

    # -----------------------------------------------------------------------------
    # Operations/Manipulations
    # -----------------------------------------------------------------------------

    def update(self,
               regulatory_events: Dict[Union[float, int], Expression] = None,
               target: 'Target' = None,
               **kwargs):

        super(Interaction, self).update(**kwargs)

        if target is not None:
            self.target: 'Target' = target

        if regulatory_events is not None:
            self.regulatory_events: Dict[Union[float, int], Expression] = regulatory_events

    def add_target(self, target: 'Target', history=True):

        if history:
            self.history.queue_command(undo_func=self.remove_target,
                                       undo_kwargs={'remove_from_model': True,
                                                    'history': False},
                                       func=self.add_target,
                                       kwargs={'target': target,
                                               'history': history})

        if self.target and not target:
            return self.remove_target(remove_from_model=True, history=False)

        elif not self.target and not target:
            return

        else:
            self._target = target
            self.target._interaction = self

            if self.model:
                self.model.add((self.target,), 'target', comprehensive=False, history=False)

                notification = Notification(content=(self,),
                                            content_type='interactions',
                                            action='add')

                self.model.notify(notification)

    def remove_target(self, remove_from_model=True, history=True):

        if history:
            self.history.queue_command(undo_func=self.add_target,
                                       undo_kwargs={'target': self.target,
                                                    'history': False},
                                       func=self.remove_target,
                                       kwargs={'remove_from_model': remove_from_model,
                                               'history': history})

        if self.target:
            target = self.target

            self.target._interaction = None
            self._target = None

            if self.model:

                if remove_from_model:
                    self.model.remove((target,), 'target', remove_orphans=False, history=False)

                notification = Notification(content=(self,),
                                            content_type='interactions',
                                            action='add')

                self.model.notify(notification)

    def add_regulatory_event(self,
                             coefficient: Union[float, int],
                             expression: Expression,
                             history=True):

        if not isinstance(expression, Expression):
            raise TypeError(f'expression must be an {Expression} object. '
                            f'To set None, provide an empty Expression()')

        if history:
            self.history.queue_command(undo_func=self.remove_regulatory_event,
                                       undo_kwargs={'coefficient': coefficient,
                                                    'expression': expression,
                                                    'remove_orphans': True,
                                                    'history': False},
                                       func=self.add_regulatory_event,
                                       kwargs={'coefficient': coefficient,
                                               'expression': expression,
                                               'history': history})

        to_add = []

        for regulator in expression.variables.values():
            regulator.update(interactions={self.id: self},
                             model=self.model)

            if self.model:
                if regulator.id not in self.model.regulators:
                    to_add.append(regulator)

        self._regulatory_events[coefficient] = expression

        if self.model:
            self.model.add(to_add, 'regulator', comprehensive=False, history=False)

            notification = Notification(content=(self,),
                                        content_type='interactions',
                                        action='add')

            self.model.notify(notification)

    def remove_regulatory_event(self,
                                coefficient: Union[float, int],
                                expression: Expression,
                                remove_orphans=True,
                                history=True):

        if not isinstance(expression, Expression):
            raise TypeError(f'expression must be an {Expression} object. '
                            f'To set None, provide an empty Expression()')

        if history:
            self.history.queue_command(undo_func=self.add_regulatory_event,
                                       undo_kwargs={'coefficient': coefficient,
                                                    'expression': expression,
                                                    'history': False},
                                       func=self.remove_regulatory_event,
                                       kwargs={'coefficient': coefficient,
                                               'expression': expression,
                                               'remove_orphans': remove_orphans,
                                               'history': history})

        to_remove = []

        for regulator in expression.variables.values():

            # noinspection PyProtectedMember
            del regulator._interactions[self.id]

            # noinspection PyProtectedMember
            if not regulator._interactions:
                to_remove.append(regulator)

        self._regulatory_events.pop(coefficient)

        if self.model:

            if remove_orphans:
                self.model.remove(to_remove, 'regulator', remove_orphans=False, history=False)

            notification = Notification(content=(self,),
                                        content_type='interactions',
                                        action='add')

            self.model.notify(notification)



import os
import re
import warnings
from collections import namedtuple
from functools import partial
from typing import Tuple

import libsbml

from mewpy.util.constants import ModelConstants
from mewpy.mew.algebra import (And, Or, Not, Less, Greater, LessEqual, GreaterEqual, BoolFalse, BoolTrue,
                           NoneAtom, Symbolic, Equal, parse_expression)


def build_symbolic(expression) -> Tuple[Symbolic, str]:
    try:

        return parse_expression(expression), ''

    except SyntaxError:

        return NoneAtom(), f'{expression} cannot be parsed. Assigning empty expression instead'


def get_sbml_doc_to_read(io):
    if isinstance(io, str):

        if os.path.exists(io):
            doc = libsbml.readSBMLFromFile(io)

        else:
            raise OSError(f'{io} is not a valid input. Provide the path or file handler')

    elif hasattr(io, 'read'):

        doc = libsbml.readSBMLFromString(io.read())

    else:

        raise OSError(f'{io} is not a valid input. Provide the path or file handler')

    return doc


def get_sbml_doc_to_write(io,
                          level,
                          version,
                          packages,
                          packages_version,
                          packages_required,
                          sbo_term=False):
    doc = None

    if isinstance(io, str):
        # string

        try:
            doc = get_sbml_doc_to_read(io)

        except OSError:

            pass

    elif hasattr(io, 'read'):

        # filepath handler

        try:
            doc = get_sbml_doc_to_read(io)

        except OSError:

            pass

    else:

        raise OSError(f'{io} is not a valid input. Provide the path or file handler')

    if doc is None:

        # only supporting SBML level 3 version 1 with fbc version 2 or qual version 1
        sbml_namespace = libsbml.SBMLNamespaces(level, version)

        for package, package_version in zip(packages, packages_version):
            sbml_namespace.addPackageNamespace(package, package_version)

        doc = libsbml.SBMLDocument(sbml_namespace)

    for package, required in zip(packages, packages_required):
        doc.setPackageRequired(package, required)

    if sbo_term:
        doc.setSBOTerm(SBO_FBA_FRAMEWORK)

    return doc


# -----------------------------------------------------------------------------
# SBML AST NODES
# -----------------------------------------------------------------------------
ASTNODE_BOOLEAN_OPERATORS = {
    libsbml.AST_LOGICAL_AND: And,
    libsbml.AST_LOGICAL_OR: Or,
    libsbml.AST_LOGICAL_NOT: Not}

ASTNODE_RELATIONAL_OPERATORS = {
    libsbml.AST_RELATIONAL_LEQ: LessEqual,
    libsbml.AST_RELATIONAL_GEQ: GreaterEqual,
    libsbml.AST_RELATIONAL_LT: Less,
    libsbml.AST_RELATIONAL_GT: Greater,
    libsbml.AST_RELATIONAL_EQ: Equal
}

ASTNODE_VALUES = {
    libsbml.AST_INTEGER,
    libsbml.AST_REAL,
    libsbml.AST_REAL_E,
    libsbml.AST_RATIONAL,
}

ASTNODE_BOOLEAN_VALUES = {
    libsbml.AST_CONSTANT_FALSE: BoolFalse(),
    libsbml.AST_CONSTANT_TRUE: BoolTrue(),
}

ASTNODE_NAME = libsbml.AST_NAME

# -----------------------------------------------------------------------------
# CONSTANTS
# -----------------------------------------------------------------------------
LOWER_BOUND_ID = 'mewpy_lb'
UPPER_BOUND_ID = 'mewpy_ub'
ZERO_BOUND_ID = 'mewpy_zero_b'

BOUND_MINUS_INF = 'minus_inf'
BOUND_PLUS_INF = 'plus_inf'

# -----------------------------------------------------------------------------
# SBO
# -----------------------------------------------------------------------------
SBO_FBA_FRAMEWORK = "SBO:0000624"
SBO_DEFAULT_FLUX_BOUND = "SBO:0000626"
SBO_FLUX_BOUND = "SBO:0000625"
SBO_EXCHANGE_REACTION = "SBO:0000627"

# -----------------------------------------------------------------------------
# UNITS
# -----------------------------------------------------------------------------
UNIT_ID = 'mmol_per_gDW_per_hr'

Unit = namedtuple('Unit', ['kind', 'scale', 'multiplier', 'exponent'])
UNITS = (Unit(kind=libsbml.UNIT_KIND_MOLE, scale=-3, multiplier=1, exponent=1),
         Unit(kind=libsbml.UNIT_KIND_GRAM, scale=0, multiplier=1, exponent=-1),
         Unit(kind=libsbml.UNIT_KIND_SECOND, scale=0, multiplier=3600, exponent=-1))

# IMPORTANT NOTE: SOME FUNCTIONS FOR PARSING SBML ENTITIES (namely, metabolic entities) HAVE BEEN HEAVILY
# INSPIRED BY THE TALENTED PEOPLE DEVELOPING COBRAPY. CHECK THE SOURCE: https://github.com/opencobra/cobrapy

# -----------------------------------------------------------------------------
# Functions for id replacements (import/export)
# -----------------------------------------------------------------------------
SBML_DOT = "__SBML_DOT__"

# -----------------------------------------------------------------------------
# Note pattern
# -----------------------------------------------------------------------------
pattern_notes = re.compile(r'<(?P<prefix>(\w+:)?)p[^>]*>(?P<content>.*?)</(?P=prefix)p>', re.IGNORECASE | re.DOTALL)

pattern_to_sbml = re.compile(r'([^0-9_a-zA-Z])')

pattern_from_sbml = re.compile(r'__(\d+)__')


def _escape_non_alphanum(non_ascii):
    """converts a non alphanumeric character to a string representation of
    its ascii number """
    return "__" + str(ord(non_ascii.group())) + "__"


def _number_to_chr(digit_str):
    """converts an ascii number to a character """
    return chr(int(digit_str.group(1)))


def _clip(sid, prefix):
    """Clips a prefix from the beginning of a string if it exists."""
    return sid[len(prefix):] if sid.startswith(prefix) else sid


def _f_gene(sid, prefix='G_'):
    """Clips gene prefix from id."""
    sid = sid.replace(SBML_DOT, ".")
    sid = pattern_from_sbml.sub(_number_to_chr, sid)
    return _clip(sid, prefix)


def _f_gene_rev(sid, prefix='G_'):
    """Adds gene prefix to id."""
    sid = pattern_to_sbml.sub(_escape_non_alphanum, sid)
    return prefix + sid.replace(".", SBML_DOT)


def _f_specie(sid, prefix='M_'):
    """Clips specie/metabolite prefix from id."""
    sid = pattern_from_sbml.sub(_number_to_chr, sid)
    return _clip(sid, prefix)


def _f_specie_rev(sid, prefix='M_'):
    """Adds specie/metabolite prefix to id."""
    sid = pattern_to_sbml.sub(_escape_non_alphanum, sid)
    return prefix + sid


def _f_reaction(sid, prefix='R_'):
    """Clips reaction prefix from id."""
    sid = pattern_from_sbml.sub(_number_to_chr, sid)
    return _clip(sid, prefix)


def _f_reaction_rev(sid, prefix='R_'):
    """Adds reaction prefix to id."""
    sid = pattern_to_sbml.sub(_escape_non_alphanum, sid)
    return prefix + sid


def _f_transition(sid, prefix='TR_'):
    """Clips transition prefix from id."""
    sid = pattern_from_sbml.sub(_number_to_chr, sid)
    return _clip(sid, prefix)


def _f_transition_rev(sid, prefix='TR_'):
    """Adds transition prefix to id."""
    sid = pattern_to_sbml.sub(_escape_non_alphanum, sid)
    return prefix + sid


F_GENE = 'F_GENE'
F_GENE_REV = 'F_GENE_REV'
F_SPECIE = 'F_SPECIE'
F_SPECIE_REV = 'F_SPECIE_REV'
F_REACTION = 'F_REACTION'
F_REACTION_REV = 'F_REACTION_REV'
F_TRANSITION = 'F_TRANSITION'
F_TRANSITION_REV = 'F_TRANSITION_REV'

F_REPLACE = {
    F_GENE: _f_gene,
    F_GENE_REV: _f_gene_rev,
    F_SPECIE: _f_specie,
    F_SPECIE_REV: _f_specie_rev,
    F_REACTION: _f_reaction,
    F_REACTION_REV: _f_reaction_rev,
    F_TRANSITION: _f_transition,
    F_TRANSITION_REV: _f_transition_rev,

}


def f_id(identifier, f_type):
    return F_REPLACE.get(f_type, lambda x: x)(identifier)


def fs_id(identifier, f_types):
    for f_type in f_types:
        identifier = f_id(identifier, f_type)

    return identifier


def add_sbml_parameter(sbml_model, parameter_id, value, constant=True, sbo=None, unit_definition=None):
    parameter = sbml_model.createParameter()
    parameter.setId(parameter_id)
    parameter.setValue(value)
    parameter.setConstant(constant)

    if sbo:
        parameter.setSBOTerm(sbo)

    if unit_definition is not None:
        parameter.setUnits(unit_definition.getId())


def get_sbml_lb_id(sbml_model, reaction, unit_definition=None):
    value = reaction.lower_bound

    if value == ModelConstants.REACTION_LOWER_BOUND:
        return LOWER_BOUND_ID

    elif value == 0:
        return ZERO_BOUND_ID

    elif value == -float('Inf'):
        return BOUND_MINUS_INF

    else:

        return None


def get_sbml_ub_id(sbml_model, reaction, unit_definition=None):
    value = reaction.upper_bound

    if value == ModelConstants.REACTION_UPPER_BOUND:
        return UPPER_BOUND_ID

    elif value == 0:
        return ZERO_BOUND_ID

    elif value == float('Inf'):
        return BOUND_PLUS_INF

    else:

        return None


def set_gpr(engine, warning, reaction, sbml_rxn_fbc):
    if not reaction.gpr.is_none:

        gpr = str(reaction.gpr)

        if gpr:

            gpr = gpr.replace('&', 'and').replace('|', 'or')

            gpa = sbml_rxn_fbc.createGeneProductAssociation()
            op = gpa.setAssociation(gpr, True, False)

            if op is None:
                engine.warnings.append(partial(warning, f"Could not set {gpr} GPR for {reaction.id}"))

            elif isinstance(op, int) and op != libsbml.LIBSBML_OPERATION_SUCCESS:
                engine.warnings.append(partial(warning, f"Could not set {gpr} GPR for {reaction.id}"))


def set_math(identifier, expression, function_term):
    math = libsbml.parseL3Formula(expression)

    op = function_term.setMath(math)

    if op is None:
        return f'Could not set {expression} expression for {identifier}'

    elif isinstance(op, int) and op != libsbml.LIBSBML_OPERATION_SUCCESS:
        return f'Could not set {expression} expression for {identifier}'

    return ''


def convert_fbc(doc):
    conversion_properties = libsbml.ConversionProperties()
    conversion_properties.addOption("convert fbc v1 to fbc v2", True, "Convert FBC-v1 model to FBC-v2")

    result = doc.convert(conversion_properties)

    if result != libsbml.LIBSBML_OPERATION_SUCCESS:
        raise ImportError("Conversion of SBML fbc plugin version 1 to 2 failed")


def write_sbml_doc(io, doc):
    if isinstance(io, str):
        libsbml.writeSBMLToFile(doc, io)

    elif hasattr(io, 'write'):
        sbml = libsbml.writeSBMLToString(doc)
        io.write(sbml)


# -----------------------------------------------------------------------------
# Warning stuff
# -----------------------------------------------------------------------------

def prom_warning(message):
    return warnings.warn(message, Warning, stacklevel=2)


def coregflux_warning(message):
    return warnings.warn(message, Warning, stacklevel=2)


def sbml_warning(message):
    return warnings.warn(message, Warning, stacklevel=2)


def cobra_warning(message):
    return warnings.warn(message, Warning, stacklevel=2)


def csv_warning(message):
    return warnings.warn(message, Warning, stacklevel=2)


def expression_warning(message):
    return warnings.warn(message, Warning, stacklevel=2)



from typing import Union, TYPE_CHECKING, Type, Dict
import sys

from mewpy.mew.algebra import Expression
from mewpy.mew.algebra import parse_expression

if TYPE_CHECKING:
    from mewpy.model import Model, MetabolicModel, RegulatoryModel
    from mewpy.mew.variables import Variable, Gene, Interaction, Metabolite, Reaction, Regulator, Target

sys.setrecursionlimit(50000)


def serialize(serialization_name, deserialization_name=None, pickle_name=None):
    def wrapper(attr):
        attr.fget.serialize = serialization_name
        attr.fget.deserialize = deserialization_name
        attr.fget.pickle = pickle_name

        return attr

    return wrapper


class Serializer:

    # -----------------------------------------------------------------------------
    # Serializing to dictionary, tuple, string and int or float
    # -----------------------------------------------------------------------------

    @staticmethod
    def _regulatory_events_serializer(regulatory_events):

        return {key: expression.to_string() for key, expression in regulatory_events.items()}

    @staticmethod
    def _expression_serializer(expression):

        return expression.to_string()

    @staticmethod
    def _obj_serializer(obj):

        if isinstance(obj, set):

            return tuple(obj)

        elif hasattr(obj, 'id'):

            return obj.id

        return obj

    @staticmethod
    def _coefficient_serializer(coefficient):

        return tuple(coefficient.coefficients)

    @staticmethod
    def _variable_container_serializer(container):

        return tuple(container.keys())

    @staticmethod
    def _key_container_serializer(container):

        return {key.id: value for key, value in container.items()}

    @staticmethod
    def _model_container_serializer(container):

        return {variable.id: variable.to_dict(serialization_format='json') for variable in container.values()}

    def _get_attribute_serializer(self, attr):

        if attr in ('id', 'name', 'types', 'aliases', 'target', 'charge', 'compartment', 'formula', 'compartments',
                    'interaction'):

            return self._obj_serializer

        elif attr == 'coefficient':

            return self._coefficient_serializer

        elif attr in ('reactions', 'regulators', 'genes', 'interactions', 'targets'):

            return self._variable_container_serializer

        elif attr == 'regulatory_events':

            return self._regulatory_events_serializer

        elif attr == 'gpr':

            return self._expression_serializer

        elif attr == 'stoichiometry':

            return self._key_container_serializer

        else:
            return lambda *args, **kwargs: {}

    def _variable_serializer(self: Union['Serializer', 'Variable', 'Model']):

        variable = {}

        for _, (attr, _, _) in self.attributes.items():
            serializer = self._get_attribute_serializer(attr=attr)

            attribute = getattr(self, attr)

            variable[attr] = serializer(attribute)

        return variable

    def _get_container_serializer(self, attr, variables=True):

        if attr in ('id', 'name', 'types'):

            return self._obj_serializer

        elif attr == 'objective':

            return self._key_container_serializer

        elif attr in ('genes', 'metabolites', 'reactions', 'interactions', 'regulators', 'targets'):

            if variables:
                return self._model_container_serializer

            return dict

        else:
            return lambda *args, **kwargs: {}

    def _model_serializer(self: Union['Serializer', 'Variable', 'Model'],
                          variables=True):

        model = {}

        for _, (attr, _, _) in self.containers.items():
            model[attr] = {}

            serializer = self._get_container_serializer(attr=attr, variables=variables)

            container = getattr(self, attr, {})

            model[attr] = serializer(container)

        return model

    # -----------------------------------------------------------------------------
    # Deserializing from dictionary, tuple, string and int or float
    # -----------------------------------------------------------------------------

    @classmethod
    def _model_deserializer(cls: Union[Type['Serializer'], Type['Variable'], Type['Model']],
                            obj,
                            variables=False):

        identifier = obj.get('id')
        types = obj.get('types')

        model = cls.from_types(types=types, identifier=identifier)

        if variables:
            children = cls._build_children(obj=obj, model=model)

        else:
            children = cls._get_children(obj=obj, model=model)

        update_attributes = {}
        for attr_name, (serialize_name, deserialize_name, _) in model.containers.items():

            if deserialize_name is not None:
                deserializer = cls._get_container_deserializer(attr=deserialize_name)

                container = obj[serialize_name]
                container = deserializer(container, children=children)

                if serialize_name in ('genes', 'metabolites', 'reactions', 'interactions', 'regulators', 'targets'):
                    setattr(model, f'_{serialize_name}', container)

                else:
                    update_attributes[deserialize_name] = container

        model.update(**update_attributes)

        return model

    @staticmethod
    def _build_children(obj, model):

        from mewpy.mew.variables import Variable

        children = {}

        for attr_name, (serialize_name, deserialize_name, _) in model.containers.items():

            if deserialize_name is not None and serialize_name in ('genes', 'metabolites', 'reactions',
                                                                   'interactions', 'regulators', 'targets'):

                container = obj[serialize_name]

                for var_id, variable in container.items():

                    if var_id in children:

                        continue

                    else:

                        children[var_id] = Variable.from_types(variable['types'],
                                                               identifier=variable['id'],
                                                               model=model)

        return children

    @staticmethod
    def _get_children(obj, model):

        children = {}

        for attr_name, (serialize_name, deserialize_name, _) in model.containers.items():

            if deserialize_name is not None and serialize_name in ('genes', 'metabolites', 'reactions',
                                                                   'interactions', 'regulators', 'targets'):

                container = obj[serialize_name]

                for var_id, variable in container.items():

                    if var_id in children:

                        continue

                    else:

                        children[var_id] = variable

        return children

    @classmethod
    def _get_container_deserializer(cls, attr):

        if attr == 'name':

            return cls._obj_deserializer

        elif attr in ('genes', 'metabolites', 'reactions',
                      'interactions', 'regulators', 'targets'):

            return cls._model_container_deserializer

        elif attr == 'objective':

            return cls._key_container_deserializer

        else:
            return lambda *args, **kwargs: {}

    @staticmethod
    def _obj_deserializer(obj, *args, **kwargs):

        return obj

    @classmethod
    def _model_container_deserializer(cls, obj, children):

        from mewpy.mew.variables import Variable

        container = {}

        for var_id, variable in obj.items():

            if isinstance(variable, Variable):

                container[var_id] = variable

            else:

                variable_dict = variable

                variable_obj = children[var_id]

                variable_attributes = cls._variable_attributes(variable_dict, variable=variable_obj, children=children)

                variable_obj.update(**variable_attributes)

                container[var_id] = variable_obj

        return container

    @staticmethod
    def _key_container_deserializer(obj, children=None, children_types=None):

        if children is False:
            return {}

        elif children is None:

            if children_types:
                from mewpy.mew.variables import Variable
                return {Variable.from_types(children_types, identifier=variable): val
                        for variable, val in obj.items()}

            else:
                from mewpy.mew.variables import Metabolite
                return {Metabolite(key): val for key, val in obj.items()}

        else:
            return {children[variable]: val for variable, val in obj.items()}

    @classmethod
    def _variable_deserializer(cls: Union[Type['Serializer'], Type['Variable'], Type['Model']],
                               obj):

        identifier = obj.get('id')
        types = obj.get('types')

        variable = cls.from_types(types=types, identifier=identifier)

        variable_attributes = cls._variable_attributes(obj, variable=variable)

        variable.update(**variable_attributes)

        return variable

    @classmethod
    def _variable_attributes(cls, obj, variable, children=None):

        variable_attributes = {}

        # Filtering attributes that must be used by the update method. Child variables associated with this variable
        # are also collected for further building
        for attr_name, (serialize_name, deserialize_name, _) in variable.attributes.items():

            if deserialize_name is not None:
                deserializer, children_types = cls._get_attribute_deserializer(attr=deserialize_name)

                attribute = obj[serialize_name]
                variable_attributes[deserialize_name] = deserializer(attribute,
                                                                     children=children,
                                                                     children_types=children_types)

        return variable_attributes

    @classmethod
    def _get_attribute_deserializer(cls, attr):

        if attr in ('name', 'aliases', 'coefficients', 'bounds', 'charge', 'compartment', 'formula'):

            return cls._obj_deserializer, None

        elif attr == 'regulatory_events':

            return cls._regulatory_events_deserializer, ['regulator']

        elif attr == 'gpr':

            return cls._expression_deserializer, ['gene']

        elif attr == 'reactions':

            return cls._variable_container_deserializer, ['reaction']

        elif attr == 'interactions':

            return cls._variable_container_deserializer, ['interaction']

        elif attr == 'stoichiometry':

            return cls._key_container_deserializer, ['metabolite']

        elif attr == 'target':

            return cls._variable_attribute_deserializer, ['target']

        elif attr == 'interaction':

            return cls._variable_attribute_deserializer, ['interaction']

        else:
            return lambda *args, **kwargs: {}, None

    @staticmethod
    def _expression_deserializer(obj, children=None, children_types=None):

        if children is None:

            symbolic = parse_expression(obj)

            from mewpy.mew.variables import Variable

            variables = {symbol.name: Variable.from_types(children_types, identifier=symbol.name)
                         for symbol in symbolic.atoms(symbols_only=True)}

            return Expression(symbolic=symbolic, variables=variables)

        else:

            symbolic = parse_expression(obj)

            variables = {symbol.name: children[symbol.name]
                         for symbol in symbolic.atoms(symbols_only=True)}

            return Expression(symbolic=symbolic, variables=variables)

    @staticmethod
    def _regulatory_events_deserializer(obj, children=None, children_types=None):

        return {state: Serializer._expression_deserializer(expression, children, children_types)
                for state, expression in obj.items()}

    @staticmethod
    def _variable_container_deserializer(obj, children=None, children_types=None):

        if children is None:

            from mewpy.mew.variables import Variable
            return {key: Variable.from_types(children_types, identifier=key) for key in obj}

        else:
            return {key: children[key] for key in obj}

    @staticmethod
    def _variable_attribute_deserializer(obj, children=None, children_types=None):

        if children is None:

            from mewpy.mew.variables import Variable
            return Variable.from_types(children_types, identifier=obj)

        else:
            return children[obj]

    # -----------------------------------------------------------------------------
    # reduce for pickle serialization
    # -----------------------------------------------------------------------------
    def __reduce__(self: Union['Serializer', 'Model', 'Variable']):

        # for further detail: https://docs.python.org/3/library/pickle.html#object.__reduce__

        from mewpy.model import Model, build_model
        from mewpy.mew.variables import Variable, build_variable

        if isinstance(self, Model):
            return build_model, (tuple(self.types), {'identifier': self.id}), self._dict_to_pickle()

        if isinstance(self, Variable):
            return build_variable, (tuple(self.types), {'identifier': self.id}), self._dict_to_pickle()

        return super(Serializer, self).__reduce__()
    # -----------------------------------------------------------------------------
    # State for pickle serialization
    # -----------------------------------------------------------------------------

    def __getstate__(self):
        return self._dict_to_pickle()

    def __setstate__(self, state):
        self.__dict__.update(state)

    # -----------------------------------------------------------------------------
    # Pickle-like serialization
    # -----------------------------------------------------------------------------

    def _pickle_variable_serializer(self: Union['Serializer', 'Variable'],
                                    to_state=True):

        attributes = {}

        for attribute, (_, _, pickle_name) in self.attributes.items():

            if pickle_name is not None:
                attribute_obj = getattr(self, pickle_name)

                if to_state:
                    attributes[pickle_name] = attribute_obj
                else:
                    attributes[attribute] = attribute_obj

        return attributes

    def _pickle_model_serializer(self: Union['Serializer', 'Model'],
                                 to_state=True):

        containers = {}

        for container, (_, _, pickle_name) in self.containers.items():

            if pickle_name is not None:
                container_obj = getattr(self, pickle_name)

                if to_state:
                    containers[pickle_name] = container_obj
                else:
                    containers[container] = container_obj

        return containers

    def _dict_to_pickle(self: Union['Serializer', 'Model', 'Variable']):

        if hasattr(self, 'containers'):
            return self._pickle_model_serializer(to_state=True)

        if hasattr(self, 'attributes'):
            return self._pickle_variable_serializer(to_state=True)

        return {}

    @classmethod
    def _pickle_variable_deserializer(cls: Union[Type['Variable']],
                                      obj):

        identifier = obj.get('id')
        types = obj.get('types')

        variable = cls.from_types(types=types, identifier=identifier)

        new_dict = {}

        for attribute, (_, _, pickle_name) in variable.attributes.items():

            if pickle_name is not None:
                new_dict[pickle_name] = obj.get(attribute)

        variable.__dict__.update(new_dict)

        return variable

    @classmethod
    def _pickle_model_deserializer(cls: Union[Type['Model']],
                                   obj):

        identifier = obj.get('id')
        types = obj.get('types')

        model = cls.from_types(types=types, identifier=identifier)

        new_dict = {}

        for container, (_, _, pickle_name) in model.containers.items():

            if pickle_name is not None:
                new_dict[pickle_name] = obj.get(container)

        model.__dict__.update(new_dict)

        return model

    # FIXME: make sure variables point to the correct model
    def to_dict(self: Union['Serializer', 'Variable', 'Model'],
                serialization_format: str = 'json',
                variables: bool = False) -> Dict[str, Union[dict,
                                                            'Gene',
                                                            'Interaction',
                                                            'Metabolite',
                                                            'Reaction',
                                                            'Regulator',
                                                            'Target']]:

        is_model = False
        if hasattr(self, 'containers'):
            is_model = True

        if serialization_format == 'pickle':

            if is_model:
                dict_obj = self._pickle_model_serializer(to_state=False)
            else:
                dict_obj = self._pickle_variable_serializer(to_state=False)

            dict_obj['types'] = self.types
            return dict_obj

        elif serialization_format == 'json':
            if is_model:
                return self._model_serializer(variables)

            return self._variable_serializer()

        return {}

    @classmethod
    def from_dict(cls: Union[Type['Serializer'], Type['Variable'], Type['Model']],
                  obj: Dict[str, Union[dict,
                                       'Gene',
                                       'Interaction',
                                       'Metabolite',
                                       'Reaction',
                                       'Regulator',
                                       'Target']],
                  serialization_format: str = 'json',
                  variables: bool = False) -> Union['Gene',
                                                    'Interaction',
                                                    'Metabolite',
                                                    'Reaction',
                                                    'Regulator',
                                                    'Target',
                                                    'MetabolicModel',
                                                    'RegulatoryModel']:

        is_model = False
        if hasattr(cls, 'containers'):
            is_model = True

        if serialization_format == 'pickle':

            if is_model:
                return cls._pickle_model_deserializer(obj)

            return cls._pickle_variable_deserializer(obj)

        elif serialization_format == 'json':

            if is_model:
                return cls._model_deserializer(obj, variables)

            return cls._variable_deserializer(obj)

        return

    # -----------------------------------------------------------------------------
    # Copy leveraging serialization
    # -----------------------------------------------------------------------------

    def __copy__(self) -> Union['Gene',
                                'Interaction',
                                'Metabolite',
                                'Reaction',
                                'Regulator',
                                'Target',
                                'MetabolicModel',
                                'RegulatoryModel']:

        obj_dict = self.to_dict(serialization_format='pickle')

        return self.from_dict(obj_dict, serialization_format='pickle')

    def copy(self) -> Union['Gene',
                            'Interaction',
                            'Metabolite',
                            'Reaction',
                            'Regulator',
                            'Target',
                            'MetabolicModel',
                            'RegulatoryModel']:

        return self.__copy__()

    def __deepcopy__(self) -> Union['Gene',
                                    'Interaction',
                                    'Metabolite',
                                    'Reaction',
                                    'Regulator',
                                    'Target',
                                    'MetabolicModel',
                                    'RegulatoryModel']:

        obj_dict = self.to_dict(variables=True)

        return self.from_dict(obj_dict, variables=True)

    def deepcopy(self) -> Union['Gene',
                                'Interaction',
                                'Metabolite',
                                'Reaction',
                                'Regulator',
                                'Target',
                                'MetabolicModel',
                                'RegulatoryModel']:

        return self.__deepcopy__()



from typing import Any, Dict, TYPE_CHECKING, Set, Union, List, Tuple, Generator

from mewpy.util.utilities import generator
from mewpy.util.serialization import serialize
from .coefficient import Coefficient
from .variable import Variable

if TYPE_CHECKING:
    from .reaction import Reaction


# TODO: methods stubs
class Gene(Variable, variable_type='gene', register=True, constructor=True, checker=True):

    def __init__(self,
                 identifier: Any,
                 coefficients: Union[Set[Union[int, float]], List[Union[int, float]], Tuple[Union[int, float]]] = None,
                 active_coefficient: Union[int, float] = None,
                 reactions: Dict[str, 'Reaction'] = None,
                 **kwargs):

        """
        A metabolic gene is regularly associated with reactions and can usually be available as
        target too.
        It holds information regarding the coefficients that can take and the reactions to which is associated


        :param identifier: identifier, e.g. b0001
        :param coefficients: the set of coefficients that this gene can take.
        These coefficients can be expanded later. 0 and 1 are added by default
        :param active_coefficient: the active coefficient
        :param reactions: the dictionary of reactions to which the gene is associated with

        """

        # the coefficient initializer sets minimum and maximum coefficients of 0.0 and 1.0
        if not coefficients:
            coefficients = (0.0, 1.0)

        if not active_coefficient:
            active_coefficient = max(coefficients)

        if not reactions:
            reactions = {}

        self._coefficient = Coefficient(variable=self, coefficients=coefficients, active=active_coefficient)
        self._reactions = reactions

        super().__init__(identifier,
                         **kwargs)

    # -----------------------------------------------------------------------------
    # Built-in
    # -----------------------------------------------------------------------------

    @property
    def types(self):

        # noinspection PyUnresolvedReferences
        _types = {Gene.variable_type}

        _types.update(super(Gene, self).types)

        return _types

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------

    @serialize('coefficient', 'coefficients', '_coefficient')
    @property
    def coefficient(self) -> Coefficient:

        if hasattr(self, '_bounds'):

            # if it is a reaction, the bounds coefficient must be returned
            return self._bounds

        elif hasattr(self, 'exchange_reaction'):

            # if it is a metabolite, the bounds coefficient of the exchange reaction must be returned
            if hasattr(self.exchange_reaction, '_bounds'):

                # noinspection PyProtectedMember
                return self.exchange_reaction._bounds

        return self._coefficient

    @serialize('reactions', 'reactions', '_reactions')
    @property
    def reactions(self) -> Dict[str, 'Reaction']:
        return self._reactions.copy()

    # -----------------------------------------------------------------------------
    # Static attributes setters
    # -----------------------------------------------------------------------------

    @reactions.setter
    def reactions(self, value: Dict[str, 'Reaction']):

        if not value:
            value = {}

        self._reactions = value

    # -----------------------------------------------------------------------------
    # Generators
    # -----------------------------------------------------------------------------

    def yield_reactions(self) -> Generator['Reaction', None, None]:

        return generator(self._reactions)

    # -----------------------------------------------------------------------------
    # Operations/Manipulations
    # -----------------------------------------------------------------------------

    def ko(self, minimum_coefficient: Union[int, float] = 0.0, history=True):

        return self.coefficient.ko(minimum_coefficient=minimum_coefficient, history=history)

    def update(self,
               coefficients: Union[Set[Union[int, float]], List[Union[int, float]], Tuple[Union[int, float]]] = None,
               active_coefficient: Union[int, float] = None,
               reactions: Dict[str, 'Reaction'] = None,
               **kwargs):

        super(Gene, self).update(**kwargs)

        if coefficients is not None:
            self.coefficient.coefficients = coefficients

        if active_coefficient is not None:
            self.coefficient.active_coefficient = active_coefficient

        if reactions is not None:
            self._reactions.update(reactions)
